#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-11-28
# @Filename: kaiju.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import warnings

from typing import TYPE_CHECKING, Optional

from jaeger import config, log
from jaeger.exceptions import JaegerError, JaegerUserWarning, TrajectoryError
from jaeger.fps import FPS
from jaeger.utils.helpers import run_in_executor


if TYPE_CHECKING:
    from matplotlib.axes import Axes

    from kaiju import RobotGridCalib


__all__ = [
    "get_robot_grid",
    "dump_robot_grid",
    "decollide",
    "get_path_pair",
    "get_snapshot",
    "unwind",
    "explode",
    "get_path_pair_in_executor",
    "decollide_in_executor",
]


def warn(message):
    warnings.warn(message, JaegerUserWarning)


def get_robot_grid(seed: int = 0, collision_buffer=None):
    """Returns a new robot grid with the destination set to the lattice position.

    If an initialised instance of the FPS is available, disabled robots will be
    set offline in the grid at their current positions.

    """

    fps = FPS.get_instance()
    if fps is None:
        warn(
            "FPS information not provided when creating the robot grid. "
            "Will not be able to disable robots."
        )

    from kaiju.robotGrid import RobotGridCalib

    kaiju_config = config["kaiju"]
    ang_step = kaiju_config["ang_step"]
    collision_buffer = collision_buffer or kaiju_config["collision_buffer"]
    alpha0, beta0 = kaiju_config["lattice_position"]
    epsilon = ang_step * kaiju_config["epsilon_factor"]

    if collision_buffer < 1.5:
        raise JaegerError("Invalid collision buffer < 1.5.")

    robot_grid = RobotGridCalib(stepSize=ang_step, epsilon=epsilon, seed=seed)
    robot_grid.setCollisionBuffer(collision_buffer)

    # TODO: This is a bit hacky. Kaiju doesn't have a collisionBuffer anymore
    # as collision buffers are per robot, but I want to keep this information
    # for when I dump and reload robot grids.
    robot_grid.collisionBuffer = collision_buffer

    if fps is not None and set(robot_grid.robotDict.keys()) != set(fps.keys()):
        raise JaegerError("Mismatch between connected positioners and robot grid.")

    for robot in robot_grid.robotDict.values():
        if fps is not None:
            positioner = fps[robot.id]
            if positioner.disabled:
                log.debug(f"Setting positioner {robot.id} offline in Kaiju.")
                robot.setAlphaBeta(positioner.alpha, positioner.beta)
                robot.setDestinationAlphaBeta(positioner.alpha, positioner.beta)
                robot.isOffline = True
                continue

        robot.setDestinationAlphaBeta(alpha0, beta0)

    return robot_grid


def dump_robot_grid(robot_grid: RobotGridCalib) -> dict:
    """Dump the information needed to restore a robot grid into a dictionary."""

    data = {}

    data["collision_buffer"] = robot_grid.collisionBuffer
    data["grid"] = {}

    for robot in robot_grid.robotDict.values():
        alpha = robot.alpha
        beta = robot.beta
        destinationAlpha = robot.destinationAlpha
        destinationBeta = robot.destinationBeta

        data["grid"][robot.id] = (alpha, beta, destinationAlpha, destinationBeta)

    return data


def load_robot_grid(data: dict, set_destination: bool = True) -> RobotGridCalib:
    """Restores a robot grid from a dump."""

    collision_buffer = data["collision_buffer"]
    robot_grid = get_robot_grid(collision_buffer=collision_buffer)

    for robot in robot_grid.robotDict.values():
        data_robot = data["grid"][robot.id]
        robot.setAlphaBeta(data_robot[0], data_robot[1])
        if set_destination:
            robot.setDestinationAlphaBeta(data_robot[2], data_robot[3])

    return robot_grid


def decollide(
    robot_grid: Optional[RobotGridCalib] = None,
    data: Optional[dict] = None,
    simple: bool = False,
) -> RobotGridCalib | dict:
    """Decollides a potentially collided grid. Raises on fail.

    Parameters
    ----------
    robot_grid
        The Kaiju ``RobotGridCalib`` instance to decollide.
    data
        A dictionary of data that can be used to reload a Kaiju robot grid
        using `.load_robot_grid`. This is useful if the function is being
        run in an executor.
    simple
        Runs ``decollideGrid()`` and returns.

    Returns
    -------
    grid
        If ``robot_grid`` is passed, returns the same grid instance after decollision.
        If ``data`` is passed, returns a dictionary describing the decollided grid
        that can be used to recreate a grid using `.load_robot_grid`.

    """

    def get_collided(robot_grid):
        collided = [rid for rid in robot_grid.robotDict if robot_grid.isCollided(rid)]
        if len(collided) == 0:
            return False
        else:
            return collided

    if robot_grid is not None and data is not None:
        raise JaegerError("robot_grid and data are mutually exclusive.")

    if data is not None:
        robot_grid = load_robot_grid(data)

    assert robot_grid is not None

    if simple:
        robot_grid.decollideGrid()
        if get_collided(robot_grid) is not False:
            raise JaegerError("Failed decolliding grid.")

        if data is not None:
            return dump_robot_grid(robot_grid)
        else:
            return robot_grid

    # First pass. If collided, decollide each robot one by one.
    # TODO: Probably this should be done in order of less to more important targets
    # to throw out the less critical ones first.
    collided = get_collided(robot_grid)
    if collided is not False:
        warn("The grid is collided. Attempting one-by-one decollision.")
        for robot_id in collided:
            if robot_grid.isCollided(robot_id):
                robot_grid.decollideRobot(robot_id)
                if robot_grid.isCollided(robot_id):
                    warn(f"Failed decolliding positioner {robot_id}.")
                else:
                    warn(f"Positioner {robot_id} was decollided.")

    # Second pass. If still collided, try a grid decollision.
    if get_collided(robot_grid) is not False:
        warn("Grid is still colliding. Attempting full grid decollision.")
        robot_grid.decollideGrid()
        if get_collided(robot_grid) is not False:
            raise JaegerError("Failed decolliding grid.")
        else:
            warn("The grid was decollided.")

    if data is not None:
        return dump_robot_grid(robot_grid)
    else:
        return robot_grid


def get_path_pair(
    robot_grid: Optional[RobotGridCalib] = None,
    data: Optional[dict] = None,
    path_generation_mode: str = "greedy",
    ignore_did_fail: bool = False,
    escape_deg: float = 5,
    speed=None,
    smooth_points=None,
    path_delay=None,
    collision_shrink=None,
) -> tuple:
    """Runs ``pathGenGreedy`` and returns the to and from destination paths.

    Parameters
    ----------
    robot_grid
        The Kaiju ``RobotGridCalib`` instance to decollide.
    data
        A dictionary of data that can be used to reload a Kaiju robot grid
        using `.load_robot_grid`. This is useful if the function is being
        run in an executor.
    path_generation_mode
        Defines the path generation algorithm to use.
        Either ``greedy`` or ``escape``.
    ignore_did_fail
        Generate paths even if path generation failed (i.e., deadlocks).
    escape_deg
        Degrees for ``pathGenEscape``.
    speed, smooth_points, path_delay, collision_shrink
        Kaiju parameters to pass to ``getPathPair``. Otherwise uses the default
        configuration values.

    Returns
    -------
    paths
        A tuple with the to destination path, from destination path, whether path
        generation failed, and the list of deadlocks

    """

    if robot_grid is not None and data is not None:
        raise JaegerError("robot_grid and data are mutually exclusive.")

    if data is not None:
        set_destination = False if path_generation_mode == "escape" else True
        robot_grid = load_robot_grid(data, set_destination=set_destination)

    assert robot_grid is not None

    if path_generation_mode == "escape":
        robot_grid.pathGenEscape(escape_deg)
        deadlocks = []
    else:
        robot_grid.pathGenGreedy()

        # Check for deadlocks.
        deadlocks = robot_grid.deadlockedRobots()
        if robot_grid.didFail and ignore_did_fail is False:
            return (None, None, robot_grid.didFail, deadlocks)

    speed = speed or config["kaiju"]["speed"]
    smooth_points = smooth_points or config["kaiju"]["smooth_points"]
    collision_shrink = collision_shrink or config["kaiju"]["collision_shrink"]
    path_delay = path_delay or config["kaiju"]["path_delay"]

    to_destination, from_destination = robot_grid.getPathPair(
        speed=speed,
        smoothPoints=smooth_points,
        collisionShrink=collision_shrink,
        pathDelay=path_delay,
    )

    return (
        to_destination,
        from_destination,
        robot_grid.didFail,
        deadlocks,
    )


async def get_path_pair_in_executor(robot_grid: RobotGridCalib, **kwargs):
    """Calls `.get_path_pair` with a process executor."""

    data = dump_robot_grid(robot_grid)
    traj_data = await run_in_executor(
        get_path_pair,
        data=data,
        executor="process",
        **kwargs,
    )

    return traj_data


async def decollide_in_executor(robot_grid: RobotGridCalib, **kwargs) -> RobotGridCalib:
    """Calls `.decollide` with a process executor."""

    data = dump_robot_grid(robot_grid)
    decollided_data = await run_in_executor(
        decollide,
        data=data,
        executor="process",
        **kwargs,
    )

    return load_robot_grid(decollided_data)


async def unwind(
    current_positions: dict[int, tuple[float | None, float | None]],
    collision_buffer: float | None = None,
    force: bool = False,
):
    """Folds all the robots to the lattice position.

    This coroutine uses a process pool executor to run Kaiju routines.

    """

    alpha0, beta0 = config["kaiju"]["lattice_position"]

    # We create the data directly since it's simple. This should be a bit faster
    # than creating a grid and dumping it.
    data = {"collision_buffer": collision_buffer, "grid": {}}
    for pid, (alpha, beta) in current_positions.items():
        data["grid"][int(pid)] = (alpha, beta, alpha0, beta0)

    (to_destination, _, did_fail, deadlocks) = await run_in_executor(
        get_path_pair,
        data=data,
        ignore_did_fail=force,
        executor="process",
    )
    if did_fail:
        if force is False:
            raise TrajectoryError(
                "Failed generating a valid unwind trajectory. "
                f"{len(deadlocks)} deadlocks were found."
            )
        else:
            log.warning("Deadlocks found in unwind but proceeding anyway.")

    return to_destination


async def explode(
    current_positions: dict[int, tuple[float | None, float | None]],
    explode_deg=20.0,
    collision_buffer: float | None = None,
):
    """Explodes the grid by a number of degrees.

    This coroutine uses a process pool executor to run Kaiju routines.

    """

    alpha0, beta0 = config["kaiju"]["lattice_position"]

    data = {"collision_buffer": collision_buffer, "grid": {}}
    for pid, (alpha, beta) in current_positions.items():
        data["grid"][int(pid)] = (alpha, beta, alpha0, beta0)

    (to_destination, *_) = await run_in_executor(
        get_path_pair,
        data=data,
        path_generation_mode="escape",
        escape_deg=explode_deg,
        ignore_did_fail=False,
        executor="process",
    )

    return to_destination


async def get_snapshot(
    fps: FPS | None = None,
    collision_buffer: float | None = None,
    highlight: int | None = None,
) -> Axes:
    """Returns matplotlib axes with the current arrangement of the FPS array."""

    fps = fps or FPS.get_instance()
    if fps.initialised is False:
        await fps.initialise()

    await fps.update_position()

    # Create a robot grid and set the current positions.
    robot_grid = get_robot_grid(collision_buffer=collision_buffer)

    for robot in robot_grid.robotDict.values():
        if robot.id not in fps.positioners.keys():
            raise ValueError(f"Positioner {robot.id} is not connected.")

        robot.setAlphaBeta(fps[robot.id].alpha, fps[robot.id].beta)

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message=".+array interface is deprecated.+")
        ax: Axes = await run_in_executor(
            robot_grid.plot_state,
            highlightRobot=highlight,
        )

    return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-11-10
# @Filename: configuration.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os
import warnings

from typing import TYPE_CHECKING, Optional, Union, cast

import numpy
import pandas
from astropy.table import Table
from astropy.time import Time
from pydl.pydlutils.yanny import write_ndarray_to_yanny

from coordio import (
    ICRS,
    Field,
    FocalPlane,
    Observed,
    PositionerApogee,
    PositionerBoss,
    Site,
    Wok,
)
from coordio.defaults import INST_TO_WAVE, POSITIONER_HEIGHT, calibration
from sdssdb.peewee.sdss5db import opsdb, targetdb

from jaeger import FPS, config
from jaeger.exceptions import JaegerError, TrajectoryError
from jaeger.kaiju import (
    decollide_in_executor,
    get_path_pair_in_executor,
    get_robot_grid,
    warn,
)

from .tools import positioner_to_wok, wok_to_positioner


if TYPE_CHECKING:
    from .design import Design


__all__ = [
    "BaseConfiguration",
    "Configuration",
    "ManualConfiguration",
    "AssignmentData",
]

PositionerType = Union[PositionerApogee, PositionerBoss]


def get_fibermap_table(length: int) -> tuple[numpy.ndarray, dict]:
    """Returns a stub for the FIBERMAP table and a default entry,"""

    fiber_map_data = [
        ("positionerId", numpy.int16),
        ("holeId", "U7"),
        ("fiberType", "U10"),
        ("assigned", numpy.int16),
        ("on_target", numpy.int16),
        ("valid", numpy.int16),
        ("xFocal", numpy.float64),
        ("yFocal", numpy.float64),
        ("alpha", numpy.float32),
        ("beta", numpy.float32),
        ("racat", numpy.float64),
        ("deccat", numpy.float64),
        ("pmra", numpy.float32),
        ("pmdec", numpy.float32),
        ("parallax", numpy.float32),
        ("ra", numpy.float64),
        ("dec", numpy.float64),
        ("lambda_eff", numpy.float32),
        ("coord_epoch", numpy.float32),
        ("spectrographId", numpy.int16),
        ("mag", numpy.dtype(("<f4", (5,)))),
        ("optical_prov", "U10"),
        ("bp_mag", numpy.float32),
        ("gaia_g_mag", numpy.float32),
        ("rp_mag", numpy.float32),
        ("h_mag", numpy.float32),
        ("catalogid", numpy.int64),
        ("carton_to_target_pk", numpy.int64),
        ("cadence", "U20"),
        ("firstcarton", "U25"),
        ("program", "U20"),
        ("category", "U20"),
        ("sdssv_boss_target0", numpy.int64),
        ("sdssv_apogee_target0", numpy.int64),
    ]

    names, formats = zip(*fiber_map_data)

    fibermap = numpy.empty((length,), dtype={"names": names, "formats": formats})

    # Define a default row with all set to "" or -999. depending on column data type.
    default = {}
    for i in range(len(names)):
        name = names[i]
        dd = numpy.dtype(formats[i])
        if name == "mag":
            value = [-999.0] * 5
        elif dd.char in ["h", "i"]:
            value = -999
        elif dd.char in ["S"]:
            value = ""
        elif dd.char in ["f", "d"]:
            value = -999.0
        else:
            value = -999.0
        default[name] = value

    default["assigned"] = 0
    default["valid"] = 0
    default["sdssv_boss_target0"] = 0
    default["sdssv_apogee_target0"] = 0

    return (fibermap, default)


class BaseConfiguration:
    """A base configuration class."""

    assignment_data: BaseAssignmentData

    def __init__(self):

        if len(calibration.positionerTable) == 0:
            raise ValueError("FPS calibrations not loaded or the array is empty.")

        # Configuration ID is None until we insert in the database.
        # Once set, it cannot be changed.
        self.configuration_id: int | None = None

        self.design: Design | None = None
        self.design_id: int | None = None

        self.fps = FPS.get_instance()

        self.robot_grid = self._initialise_grid()

    def _initialise_grid(self):

        self.robot_grid = get_robot_grid()

        return self.robot_grid

    def __repr__(self):
        return f"<Configuration (configuration_id={self.configuration_id}>"

    async def get_trajectory(
        self,
        decollide: bool = False,
        simple_decollision: bool = False,
    ):
        """Returns a trajectory dictionary from the folded position."""

        assert isinstance(self, BaseConfiguration)

        # Just to be sure, reinitialise the grid.
        self.robot_grid = self._initialise_grid()

        ftable = self.assignment_data.fibre_table
        alpha0, beta0 = config["kaiju"]["lattice_position"]

        for robot in self.robot_grid.robotDict.values():
            if robot.id not in ftable.index.get_level_values(0):
                raise JaegerError(f"Positioner {robot.id} is not assigned.")

            # Get the first of the three fibres since all have the same alpha, beta.
            rdata = ftable.loc[robot.id].iloc[0]
            if rdata.valid:
                robot.setAlphaBeta(rdata.alpha, rdata.beta)
                robot.setDestinationAlphaBeta(alpha0, beta0)
                continue
            raise JaegerError(f"Positioner {robot.id} has no valid coordinates.")

        for r in self.robot_grid.robotDict.values():
            cols = ["xwok_kaiju", "ywok_kaiju", "zwok_kaiju"]
            ftable.loc[(r.id, "APOGEE"), cols] = r.apWokXYZ
            ftable.loc[(r.id, "BOSS"), cols] = r.bossWokXYZ
            ftable.loc[(r.id, "Metrology"), cols] = r.metWokXYZ

        if decollide:
            # TODO: if this is run, the configuration will change and we don't know
            # where positioners are placed.
            await decollide_in_executor(self.robot_grid, simple=simple_decollision)

        result = await get_path_pair_in_executor(self.robot_grid)
        _, from_destination, did_fail, deadlocks = result
        if did_fail:
            raise TrajectoryError(
                "Failed generating a valid trajectory. "
                f"{len(deadlocks)} deadlocks were found."
            )

        return from_destination

    @property
    def ingested(self):
        """Returns `True` if the configuration has been loaded to opsdb."""

        if self.configuration_id is None:
            return False

        return (
            opsdb.Configuration.select()
            .where(opsdb.Configuration.configuration_id == self.configuration_id)
            .exists()
        )

    def write_to_database(self, replace=False):
        """Writes the configuration to the database."""

        assert self.assignment_data.site.time
        epoch = self.assignment_data.site.time.jd

        if self.configuration_id is None:

            with opsdb.database.atomic():

                if self.design is None:
                    if (
                        targetdb.Design()
                        .select()
                        .where(targetdb.Design.design_id == self.design_id)
                        .exists()
                    ):
                        design_id = self.design_id
                    else:
                        design_id = None
                else:
                    design_id = self.design.design_id

                configuration = opsdb.Configuration(
                    configuration_id=self.configuration_id,
                    design_id=design_id,
                    epoch=epoch,
                    calibration_version=calibration.fps_calibs_version,
                )
                configuration.save()

            if configuration.configuration_id is None:
                raise JaegerError("Failed loading configuration.")

            self.configuration_id = configuration.configuration_id

        else:

            if self.configuration_id is None:
                raise JaegerError("Must have a configuration_id to replace.")

            if not self.ingested:
                raise JaegerError(
                    f"Configuration ID {self.configuration_id} does not exists "
                    "in opsdb and cannot be replaced."
                )

            if replace is False:
                raise JaegerError(
                    f"Configuration ID {self.configuration_id} has already "
                    "been loaded. Use replace=True to overwrite it."
                )

            with opsdb.database.atomic():
                opsdb.AssignmentToFocal.delete().where(
                    opsdb.AssignmentToFocal.configuration_id == self.configuration_id
                ).execute(opsdb.database)

                configuration = opsdb.Configuration.get_by_id(self.configuration_id)
                configuration.delete_instance()
                self.configuration_id = None

                return self.write_to_database(replace=False)

        a_data = self.assignment_data.fibre_table

        focals = []
        for index, data in a_data.iterrows():

            index = cast(tuple, index)
            pid = index[0]
            hole_id = data.hole_id
            try:
                if data.valid == 0:
                    raise ValueError(f"Invalid coordinate found for positioner {pid}.")

                xfocal, yfocal = data.xfocal, data.yfocal
                if xfocal == -999.0 or yfocal == -999.0:
                    xfocal = yfocal = None

            except ValueError:
                xfocal = yfocal = None

            if self.design and hole_id in self.design.target_data:
                assignment_pk = self.design.target_data[hole_id]["assignment_pk"]
            else:
                assignment_pk = None

            focals.append(
                dict(
                    assignment_pk=assignment_pk,
                    xfocal=xfocal,
                    yfocal=yfocal,
                    positioner_id=pid,
                    configuration_id=self.configuration_id,
                )
            )

        with opsdb.database.atomic():
            opsdb.AssignmentToFocal.insert_many(focals).execute(opsdb.database)

    def write_summary(self, overwrite=False):
        """Writes the confSummary file."""

        # TODO: some time may be saved by doing a single DB query and retrieving
        # all the info at once for all the assignments. Need to be careful
        # to maintain the order.

        if self.configuration_id is None:
            raise JaegerError("Configuration needs to be set and loaded to the DB.")

        adata = self.assignment_data
        fdata = self.assignment_data.fibre_table

        time = Time.now()

        design = self.design

        header = {
            "configuration_id": self.configuration_id,
            "targeting_version": -999,
            "robostrategy_run": "NA",
            "fps_calibrations_version": calibration.fps_calibs_version,
            "design_id": self.design_id,
            "field_id": -999,
            "instruments": "BOSS APOGEE",
            "epoch": adata.site.time.jd if adata.site.time else -999,
            "obstime": time.strftime("%a %b %d %H:%M:%S %Y"),
            "MJD": int(time.mjd),  # TODO: this should be SJD
            "observatory": adata.observatory,
            "temperature": -999,  # TODO
            "raCen": -999.0,
            "decCen": -999.0,
        }

        if design:
            header.update(
                {
                    "robostrategy_run": design.field["rs_run"],
                    "field_id": design.field["field_id"],
                    "raCen": design.field["racen"],
                    "decCen": design.field["deccen"],
                }
            )

        fibermap, default = get_fibermap_table(len(fdata))

        i = 0
        for index, row_data in fdata.iterrows():

            index = cast(tuple, index)

            pid, fibre_type = index
            hole_id = row_data.hole_id

            # Start with the default row.
            row = default.copy()

            if fibre_type.upper() == "APOGEE":
                spec_id = 0
            elif fibre_type.upper() == "BOSS":
                spec_id = 1
            else:
                spec_id = -1

            # Update data that is valid for all fibres.
            row.update(
                {
                    "positionerId": pid,
                    "holeId": hole_id,
                    "fiberType": fibre_type.upper(),
                    "assigned": row_data.assigned,
                    "valid": row_data.valid,
                    "on_target": row_data.on_target,
                    "xFocal": row_data.xfocal,
                    "yFocal": row_data.yfocal,
                    "alpha": row_data.alpha,
                    "beta": row_data.beta,
                    "ra": row_data.ra_epoch,
                    "dec": row_data.ra_epoch,
                    "spectrographId": spec_id,
                }
            )

            # And now only the one that is associated with a target.
            if row_data.assigned == 1 and design and hole_id in design.target_data:
                target = design.target_data[hole_id]
                row.update(
                    {
                        "racat": target["ra"],
                        "deccat": target["dec"],
                        "pmra": target["pmra"] or -999.0,
                        "pmdec": target["pmdec"] or -999.0,
                        "parallax": target["parallax"] or -999.0,
                        "coord_epoch": target["epoch"] or -999.0,
                        "lambda_eff": target["lambda_eff"] or -999.0,
                        "catalogid": target["catalogid"] or -999.0,
                        "carton_to_target_pk": target["carton_to_target_pk"] or -999.0,
                        "cadence": target["cadence"] or "",
                        "firstcarton": target["carton"] or "",
                        "program": target["program"] or "",
                        "category": target["category"] or "",
                    }
                )

                optical_mag = [target[m] or -999.0 for m in ["g", "r", "i", "z"]]
                optical_mag = [-999.0] + optical_mag  # u band
                row.update(
                    {
                        "mag": optical_mag,
                        "optical_prov": target["optical_prov"] or "",
                        "bp_mag": target["bp"] or -999.0,
                        "gaia_g_mag": target["gaia_g"] or -999.0,
                        "rp_mag": target["rp"] or -999.0,
                        "h_mag": target["h"] or -999.0,
                    }
                )

            fibermap[i] = tuple(row.values())

            i += 1

        fibermap = Table(fibermap)
        fibermap.sort(["positionerId", "fiberType"])

        if "SDSSCORE_DIR" not in os.environ:
            raise JaegerError("$SDSSCORE_DIR is not set. Cannot write summary file.")

        sdsscore_dir = os.environ["SDSSCORE_DIR"]
        path = os.path.join(
            sdsscore_dir,
            self.assignment_data.observatory.lower(),
            "summary_files",
            f"{int(self.configuration_id / 100):04d}XX",
            f"confSummary-{self.configuration_id}.par",
        )

        if os.path.exists(path):
            if overwrite:
                warn(f"Summary file {os.path.basename(path)} exists. Overwriting it.")
                os.remove(path)
            else:
                raise JaegerError(f"Summary file {os.path.basename(path)} exists.")

        os.makedirs(os.path.dirname(path), exist_ok=True)

        write_ndarray_to_yanny(
            path,
            [fibermap],
            structnames=["FIBERMAP"],
            hdr=header,
            enums={"fiberType": ("FIBERTYPE", ("BOSS", "APOGEE", "METROLOGY", "NONE"))},
        )


class Configuration(BaseConfiguration):
    """A configuration based on a target design."""

    assignment_data: AssignmentData

    def __init__(self, design: Design, **kwargs):

        super().__init__(**kwargs)

        self.design = design
        self.design_id = design.design_id
        self.assignment_data = AssignmentData(self)

    def __repr__(self):
        return (
            f"<Configuration (configuration_id={self.configuration_id} "
            f"design_id={self.design_id})>"
        )

    def recompute_coordinates(self, jd: Optional[float] = None):
        """Recalculates the coordinates. ``jd=None`` uses the current time."""

        if self.configuration_id is not None:
            raise JaegerError(
                "Cannot recompute coordinates once the configuration "
                "has been loaded to the database."
            )

        self.assignment_data.compute_coordinates(jd=jd)

        assert self.assignment_data.site.time
        self.epoch = self.assignment_data.site.time.jd


class ManualConfiguration(BaseConfiguration):
    """A configuration create manually.

    Parameters
    ----------
    target_data
        A dictionary containing the targeting information. It must be a
        mapping of hole ID to dictionary. The hole ID dictionaries must
        include one of the following pairs: ``(ra_icrs, dec_icrs)``,
        ``(xwok, ywok)``, or ``(alpha, beta)`` if order of priority.
        The remaining coordinates will be filled out using coordinate
        transformations. If ``ra_icrs/dec_icrs`` are passed, the column
        ``epoch`` is also required. An additional key, ``fibre_type``
        must be set for each target with value ``'APOGEE'``, ``'BOSS'``,
        or ``'Metrology``.
    field_centre
        A tuple or array with the boresight coordinates as current epoch
        RA/Dec. If `None`, target coordinates must be positioner or wok.
    design_id
        A design identifier for this configuration.
    observatory
        The observatory name. If `None`, uses the value from the configuration.
    position_angle
        The position angle of the field.

    """

    assignment_data: ManualAssignmentData

    def __init__(
        self,
        target_data: dict[str, dict],
        field_centre: tuple[float, float] | numpy.ndarray | None = None,
        design_id: int = -999,
        observatory: str | None = None,
        position_angle: float = 0.0,
    ):

        super().__init__()

        self.design = None
        self.design_id = design_id
        self.epoch = None

        if observatory is None:
            if config["observatory"] != "${OBSERATORY}":
                observatory = config["observatory"]
                assert isinstance(observatory, str)
            else:
                raise ValueError("Unknown site.")

        self.target_data = target_data
        self.assignment_data = ManualAssignmentData(
            self,
            target_data,
            observatory,
            field_centre=field_centre,
            position_angle=position_angle,
        )

    @classmethod
    def create_folded(cls, **kwargs):
        """Creates a folded configuration."""

        positionerTable = calibration.positionerTable.reset_index()
        alphaL, betaL = config["kaiju"]["lattice_position"]
        data = {
            positionerTable.iloc[i].holeID: {"alpha": alphaL, "beta": betaL}
            for i in range(len(positionerTable))
        }

        return cls(data, **kwargs)


class BaseAssignmentData:
    """Information about the target assignment along with coordinate transformation."""

    boresight: Observed
    position_angle: float

    _columns = [
        ("positioner_id", numpy.int32, None),
        ("fibre_type", "U10", None),
        ("hole_id", "U10", ""),
        ("assigned", numpy.int8, 0),
        ("valid", numpy.int8, 1),
        ("on_target", numpy.int8, 0),
        ("disabled", numpy.int8, 0),
        ("offline", numpy.int8, 0),
        ("deadlocked", numpy.int8, 0),
        ("collided", numpy.int8, 0),
        ("wavelength", numpy.float32, -999.0),
        ("ra_icrs", numpy.float64, -999.0),
        ("dec_icrs", numpy.float64, -999.0),
        ("ra_epoch", numpy.float64, -999.0),
        ("dec_epoch", numpy.float64, -999.0),
        ("alt", numpy.float64, -999.0),
        ("az", numpy.float64, -999.0),
        ("xfocal", numpy.float64, -999.0),
        ("yfocal", numpy.float64, -999.0),
        ("xwok", numpy.float64, -999.0),
        ("ywok", numpy.float64, -999.0),
        ("zwok", numpy.float64, -999.0),
        ("xwok_kaiju", numpy.float64, -999.0),
        ("ywok_kaiju", numpy.float64, -999.0),
        ("zwok_kaiju", numpy.float64, -999.0),
        ("xwok_measured", numpy.float64, -999.0),
        ("ywok_measured", numpy.float64, -999.0),
        ("xtangent", numpy.float64, -999.0),
        ("ytangent", numpy.float64, -999.0),
        ("ztangent", numpy.float64, -999.0),
        ("alpha", numpy.float64, -999.0),
        ("beta", numpy.float64, -999.0),
    ]

    def __init__(
        self,
        configuration: Configuration | ManualConfiguration,
        observatory: Optional[str] = None,
    ):

        self.configuration = configuration

        self.design = configuration.design
        self.design_id = self.configuration.design_id

        self.observatory: str
        if observatory is None:
            if self.design is None:
                raise ValueError("Cannot determine observatory.")
            self.observatory = self.design.field["observatory"].upper()
        else:
            self.observatory = observatory

        self.site = Site(self.observatory)
        self.site.set_time()

        positionerTable = calibration.positionerTable
        wokCoords = calibration.wokCoords
        if positionerTable is None or wokCoords is None:
            raise ValueError("FPS calibrations not loaded.")

        self.wok_data = pandas.merge(
            positionerTable.reset_index(),
            wokCoords.reset_index(),
            on="holeID",
        )
        self.wok_data.set_index("positionerID", inplace=True)

        if self.design:
            self.target_data = self.design.target_data
            self.position_angle = self.design.field["position_angle"]
        else:
            self.target_data = {}

        names, _, values = zip(*self._columns)
        self._defaults = {
            name: values[i] for i, name in enumerate(names) if values[i] is not None
        }

        self.fibre_table: pandas.DataFrame

    def __repr__(self):
        return f"<{self.__class__.__name__} (design_id={self.design_id})>"

    def _create_fibre_table(self):
        """Creates an empty fibre table."""

        names, dtypes, defaults = zip(*self._columns)

        # Create empty dataframe with zero values. Fill out all the index data.
        npositioner = len(self.wok_data)
        base = numpy.zeros((npositioner * 3,), dtype=list(zip(names, dtypes)))

        for i in range(len(defaults)):
            if defaults[i] is None:
                continue
            base[names[i]] = defaults[i]

        i = 0
        for pid in self.wok_data.index.tolist():
            for ft in ["APOGEE", "BOSS", "Metrology"]:
                base["positioner_id"][i] = pid
                base["fibre_type"][i] = ft
                base["hole_id"][i] = self.wok_data.loc[pid].holeID
                i += 1

        fibre_table = pandas.DataFrame(base)

        fibre_table.fibre_type = fibre_table.fibre_type.astype("category")
        fibre_table.hole_id = fibre_table.hole_id.astype("string")

        fibre_table.set_index(["positioner_id", "fibre_type"], inplace=True)
        fibre_table = fibre_table.sort_index()

        return fibre_table

    def _from_icrs(self):
        """Loads fibre data from target data using ICRS coordinates."""

        kaiju_config = config["kaiju"]
        alpha0, beta0 = kaiju_config["lattice_position"]

        data = {}
        for pid in self.wok_data.index:

            positioner_data = self.wok_data.loc[pid]
            hole_id = positioner_data.holeID

            target_fibre_type: str | None = None
            if hole_id in self.target_data:

                # First do the assigned fibre.
                ftype = self.target_data[hole_id]["fibre_type"].upper()
                target_fibre_type = ftype

                target_data = self.target_data[hole_id]
                positioner_data = self.icrs_to_positioner(
                    pid,
                    ftype,
                    target_data,
                    position_angle=self.position_angle,
                    update=False,
                    on_target=1,
                    assigned=1,
                )
                data[(pid, ftype)] = positioner_data

            else:

                # If a positioner does not have an assigned target, leave it folded.
                target_fibre_type = None
                positioner_data = {"alpha": alpha0, "beta": beta0}

            # Now calculate some coordinates for the other two non-assigned fibres.
            for ftype in ["APOGEE", "BOSS", "Metrology"]:
                if ftype == target_fibre_type:
                    continue

                icrs_data = self.positioner_to_icrs(
                    pid,
                    ftype,
                    positioner_data["alpha"],
                    positioner_data["beta"],
                    position_angle=self.position_angle,
                    update=False,
                )
                data[(pid, ftype)] = icrs_data

        # Now do a single update of the whole fibre table.
        self.fibre_table.update(pandas.DataFrame.from_dict(data, orient="index"))

    def validate(self):
        """Validates the fibre table."""

        alpha_beta = self.fibre_table[["alpha", "beta"]]
        na = alpha_beta.isna().any(axis=1)
        over_180 = self.fibre_table.beta > 180

        self.fibre_table.loc[na | over_180, "valid"] = 0

    def icrs_to_positioner(
        self,
        positioner_id: int,
        fibre_type: str,
        target_data: dict,
        position_angle: float = 0.0,
        update: bool = True,
        **kwargs,
    ):
        """Converts from ICRS coordinates."""

        hole_id = self.wok_data.loc[positioner_id].holeID
        wavelength = INST_TO_WAVE.get(fibre_type.capitalize(), INST_TO_WAVE["GFA"])

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)

            ra = target_data["ra"]
            dec = target_data["dec"]
            pmra = target_data["pmra"]
            pmdec = target_data["pmdec"]
            parallax = target_data["parallax"]
            epoch = target_data["epoch"]

            icrs = ICRS(
                [[ra, dec]],
                pmra=numpy.nan_to_num(pmra, nan=0),
                pmdec=numpy.nan_to_num(pmdec, nan=0),
                parallax=numpy.nan_to_num(parallax),
                epoch=epoch,
            )

            assert self.site.time
            icrs_epoch = icrs.to_epoch(self.site.time.jd, site=self.site)

            observed = Observed(icrs, wavelength=wavelength, site=self.site)
            field = Field(observed, field_center=self.boresight)
            focal = FocalPlane(field, wavelength=wavelength, site=self.site)
            wok = Wok(focal, site=self.site, obsAngle=position_angle)

            positioner, tangent = wok_to_positioner(
                hole_id,
                self.site.name,
                fibre_type,
                wok[0][0],
                wok[0][1],
            )

        row = self._defaults.copy()
        row.update(
            {
                "hole_id": hole_id,
                "wavelength": INST_TO_WAVE[fibre_type.capitalize()],
                "ra_icrs": icrs[0, 0],
                "dec_icrs": icrs[0, 1],
                "ra_epoch": icrs_epoch[0, 0],
                "dec_epoch": icrs_epoch[0, 1],
                "alt": observed[0, 0],
                "az": observed[0, 1],
                "xfocal": focal[0, 0],
                "yfocal": focal[0, 1],
                "xwok": wok[0, 0],
                "ywok": wok[0, 1],
                "zwok": wok[0, 2],
                "xtangent": tangent[0],
                "ytangent": tangent[1],
                "ztangent": tangent[2],
                "alpha": positioner[0],
                "beta": positioner[1],
            }
        )
        row.update(kwargs)

        if update:
            self.fibre_table.loc[(positioner_id, fibre_type)] = pandas.Series(row)

        return row

    def positioner_to_icrs(
        self,
        positioner_id: int,
        fibre_type: str,
        alpha: float,
        beta: float,
        position_angle: float = 0.0,
        update: bool = True,
        **kwargs,
    ):
        """Converts from positioner to ICRS coordinates."""

        wavelength = INST_TO_WAVE.get(fibre_type.capitalize(), INST_TO_WAVE["GFA"])

        assert self.site.time

        hole_id = self.wok_data.at[positioner_id, "holeID"]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)

            wok, _ = positioner_to_wok(
                hole_id,
                self.site.name,
                fibre_type,
                alpha,
                beta,
            )

            focal = FocalPlane(
                Wok([wok], site=self.site, obsAngle=position_angle),
                wavelength=wavelength,
                site=self.site,
            )

            field = Field(focal, field_center=self.boresight)
            obs = Observed(field, site=self.site, wavelength=wavelength)
            icrs = ICRS(obs, epoch=self.site.time.jd)

        row = self._defaults.copy()
        row.update(
            {
                "hole_id": hole_id,
                "wavelength": wavelength,
                "ra_epoch": icrs[0, 0],
                "dec_epoch": icrs[0, 1],
                "xfocal": focal[0, 0],
                "yfocal": focal[0, 1],
                "xwok": wok[0],
                "ywok": wok[1],
                "zwok": wok[2],
                "alpha": alpha,
                "beta": beta,
            }
        )
        row.update(kwargs)

        if update:
            self.fibre_table.loc[(positioner_id, fibre_type)] = pandas.Series(row)

        return row


class AssignmentData(BaseAssignmentData):
    """Assignment data from a valid design with associated target information."""

    design: Design

    def __init__(self, configuration: Configuration):

        super().__init__(configuration)

        self.compute_coordinates()

    def compute_coordinates(self, jd: Optional[float] = None):
        """Computes coordinates in different systems."""

        self.fibre_table = self._create_fibre_table()
        self.site.set_time(jd)

        icrs_bore = ICRS([[self.design.field["racen"], self.design.field["deccen"]]])
        self.boresight = Observed(
            icrs_bore,
            site=self.site,
            wavelength=INST_TO_WAVE["GFA"],
        )

        # Load fibre data using ICRS coordinates.
        self._from_icrs()

        # Final validation
        self.validate()


class ManualAssignmentData(BaseAssignmentData):
    """Assignment data from a manual configuration.

    Parameters
    ----------
    configuration
        The parent `.ManualConfiguration`.
    target_data
        A dictionary containing the targeting information. It must be a
        mapping of hole ID to dictionary. The hole ID dictionaries must
        include one of the following pairs: ``(ra_icrs, dec_icrs)``,
        ``(xwok, ywok)``, or ``(alpha, beta)`` if order of priority.
        The remaining coordinates will be filled out using coordinate
        transformations. If ``ra_icrs/dec_icrs`` are passed, an additional
        column ``epoch`` is required.
    observatory
        The observatory name. If `None`, uses the value from the configuration.
    field_centre
        A tuple or array with the boresight coordinates as current epoch
        RA/Dec. If `None`, target data must be positioner or wok coordinates.
    position_angle
        The position angle of the field.

    """

    boresight: Observed | None = None

    def __init__(
        self,
        configuration: ManualConfiguration,
        target_data: dict,
        observatory: str,
        field_centre: tuple[float, float] | numpy.ndarray | None = None,
        position_angle: float = 0.0,
    ):

        super().__init__(configuration, observatory=observatory)

        self.target_data = target_data

        if field_centre is not None:
            self.field_centre = numpy.array(field_centre)
        else:
            self.field_centre = None

        self.position_angle = position_angle

        self.compute_coordinates()

    def compute_coordinates(self, jd: Optional[float] = None):
        """Computes coordinates in different systems."""

        self.fibre_table = self._create_fibre_table()
        self.site.set_time(jd)

        if self.field_centre:
            icrs_bore = ICRS([self.field_centre])
            self.boresight = Observed(
                icrs_bore,
                site=self.site,
                wavelength=INST_TO_WAVE["GFA"],
            )

        if len(self.target_data) == 0:
            raise ValueError("Target data has zero length.")

        sample_target = list(self.target_data.values())[0]

        if "fibre_type" not in sample_target:
            raise ValueError("fibre_type column missing from target data.")

        if "ra_icrs" in sample_target and "dec_icrs" in sample_target:
            if "epoch" not in sample_target:
                raise ValueError("Missing epoch information.")
            if self.boresight is None:
                raise ValueError(
                    "Creating a manual configuration from ICRS "
                    "coordinates requires defining a field centre."
                )
            self._from_icrs()
        elif "xwok" in sample_target and "ywok" in sample_target:
            self._from_wok()
        elif "alpha" in sample_target and "beta" in sample_target:
            self._from_positioner()
        else:
            raise ValueError("Target data does not contain the necessary columns.")

        # Final validation.
        self.validate()

    def _from_wok(self):
        """Loads fibre data from target data using wok coordinates."""

        self._check_all_assigned()

        # Calculate positioner coordinates for each wok coordinate.
        for hole_id, data in self.target_data.items():
            xwok = data["xwok"]
            ywok = data["ywok"]
            zwok = data.get("zwok", POSITIONER_HEIGHT)
            fibre_type = data["fibre_type"]

            (alpha, beta), _ = wok_to_positioner(
                hole_id,
                self.site.name,
                fibre_type,
                xwok,
                ywok,
                zwok=zwok,
            )

            self.target_data[hole_id].update({"alpha": alpha, "beta": beta})

        # Now simply call _from_positioner()
        self._from_positioner()

        # We want to keep the original wok coordinates that now have been overridden
        # by calling _from_positioner.
        wok_coords = []
        for idx in range(len(self.fibre_table)):
            row: pandas.Series = self.fibre_table.iloc[idx]
            hole_id = row.hole_id
            assigned = row.assigned

            if assigned == 1 and hole_id in self.target_data:
                target = self.target_data[hole_id]
                wok_coords.append(
                    [
                        target["xwok"],
                        target["ywok"],
                        target.get("zwok", -999.0),
                    ]
                )
            else:
                wok_coords.append([-999.0] * 3)
        self.fibre_table[["xwok", "ywok", "zwok"]] = wok_coords

    def _from_positioner(self):
        """Loads fibre data from target data using positioner coordinates."""

        self._check_all_assigned()

        data = {}
        for pid in self.wok_data.index:

            hole_id = self.wok_data.at[pid, "holeID"]
            alpha = self.target_data[hole_id]["alpha"]
            beta = self.target_data[hole_id]["beta"]

            # Now calculate some coordinates for the other two non-assigned fibres.
            for ftype in ["APOGEE", "BOSS", "Metrology"]:

                assigned = self.target_data[hole_id]["fibre_type"] == ftype

                # If boresight has been set, go all the way from positioner to ICRS.
                # Otherwise just calculate tangent and wok.
                if self.boresight:
                    icrs_data = self.positioner_to_icrs(
                        pid,
                        ftype,
                        alpha,
                        beta,
                        position_angle=self.position_angle,
                        update=False,
                        assigned=assigned,
                        on_target=assigned,
                    )
                    data[(pid, ftype)] = icrs_data

                else:

                    wok, tangent = positioner_to_wok(
                        hole_id,
                        self.site.name,
                        ftype,
                        alpha,
                        beta,
                    )

                    row = self._defaults.copy()
                    row.update(
                        {
                            "hole_id": hole_id,
                            "alpha": alpha,
                            "beta": beta,
                            "xwok": wok[0],
                            "ywok": wok[1],
                            "zwok": wok[2],
                            "xtangent": tangent[0],
                            "ytangent": tangent[1],
                            "ztangent": tangent[2],
                            "assigned": assigned,
                            "on_target": assigned,
                        }
                    )
                    data[(pid, ftype)] = row

        # Now do a single update of the whole fibre table.
        self.fibre_table.update(pandas.DataFrame.from_dict(data, orient="index"))

    def _check_all_assigned(self):
        """Check that all the positioners are in ``target_data``."""

        if len(set(self.target_data) - set(self.wok_data.holeID)) > 0:
            raise ValueError("Not all the positioners have been assigned a target.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-11-12
# @Filename: cli.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import logging
import os
import signal
import socket
import sys
import warnings
from functools import wraps

from typing import Any, Optional

import click
import numpy
from click_default_group import DefaultGroup

from sdsstools.daemonizer import DaemonGroup

from jaeger import can_log, config, log
from jaeger.commands.bootloader import load_firmware
from jaeger.commands.calibration import calibrate_positioner
from jaeger.exceptions import JaegerError, JaegerUserWarning
from jaeger.fps import FPS, LOCK_FILE
from jaeger.positioner import Positioner
from jaeger.testing import VirtualFPS


__FPS__ = None


def shutdown(sign):
    """Shuts down the FPS and stops the positioners in case of a signal interrupt."""

    if __FPS__ is not None:
        __FPS__.send_command("SEND_TRAJECTORY_ABORT", positioner_ids=None, now=True)
        log.error(f"stopping positioners and cancelling due to {sign.name}")
        sys.exit(0)
    else:
        log.error(f"cannot shutdown FPS before {sign.name}")
        sys.exit(1)


def cli_coro(f):
    """Decorator function that allows defining coroutines with click."""

    @wraps(f)
    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)
        for ss in signals:
            loop.add_signal_handler(ss, shutdown, ss)
        return loop.run_until_complete(f(*args, **kwargs))

    return wrapper


class FPSWrapper(object):
    """A helper to store FPS initialisation parameters."""

    def __init__(
        self,
        profile,
        ieb=None,
        initialise=True,
        npositioners=10,
        enable_low_temperature=True,
        use_lock=True,
    ):

        self.profile = profile
        if self.profile in ["test", "virtual"]:
            self.profile = "virtual"

        self.ieb = ieb
        self.enable_low_temperature = enable_low_temperature
        self.initialise = initialise
        self.use_lock = use_lock

        self.vpositioners = []
        self.npositioners = npositioners

        self.fps = None

    async def __aenter__(self):

        global __FPS__

        # If profile is test we start a VirtualFPS first so that it can respond
        # to the FPS class.
        if self.profile == "virtual":
            self.fps = VirtualFPS()
            for pid in range(self.npositioners):
                self.fps.add_virtual_positioner(pid + 1)
        else:
            self.fps = FPS(can=self.profile, ieb=self.ieb)

        __FPS__ = self.fps

        if self.initialise:
            await self.fps.initialise(
                enable_low_temperature=self.enable_low_temperature,
                use_lock=self.use_lock,
            )
            if self.enable_low_temperature is False:
                warnings.warn(
                    "Disabling low temperature handling for sextant.",
                    JaegerUserWarning,
                )

        return self.fps

    async def __aexit__(self, *excinfo):
        try:
            if self.fps is None:
                return
            await self.fps.shutdown()
        except JaegerError as err:
            warnings.warn(f"Failed shutting down FPS: {err}", JaegerUserWarning)


pass_fps = click.make_pass_decorator(FPSWrapper, ensure=True)


@click.group(cls=DefaultGroup, default="actor", default_if_no_args=True)
@click.option(
    "-c",
    "--config",
    "config_file",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to the user configuration file.",
)
@click.option(
    "-p",
    "--profile",
    type=str,
    help="The bus interface profile.",
)
@click.option(
    "--virtual",
    is_flag=True,
    help="Runs a virtual FPS with virtual positioners. Same as --profile=virtual.",
)
@click.option(
    "-n",
    "--npositioners",
    type=int,
    default=10,
    help="How many virtual positioners must be connected to the virtual FPS.",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Debug mode. Use additional v for more details.",
)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help="Disable all console logging.",
)
@click.option(
    "--ieb/--no-ieb",
    default=None,
    help="Does not connect to the IEB.",
)
@click.option(
    "-s",
    "--sextant",
    is_flag=True,
    help="Use engineering sextant instead of IEB. "
    "Modifies the internal configuration file.",
)
@click.option(
    "--allow-host",
    is_flag=True,
    help="Allows running jager in a host other than sdss5-fps.",
)
@click.option(
    "--no-lock",
    is_flag=True,
    help="Do not use the lock file, or ignore it if present.",
)
@click.pass_context
def jaeger(
    ctx,
    config_file,
    profile,
    verbose,
    quiet,
    ieb,
    sextant,
    virtual,
    npositioners,
    allow_host,
    no_lock,
):
    """CLI for the SDSS-V focal plane system.

    If called without subcommand starts the actor.

    """

    if allow_host is False:
        hostname = socket.getfqdn()
        if hostname.endswith("apo.nmsu.edu") or hostname.endswith("lco.cl"):
            if not hostname.startswith("sdss5-fps"):
                raise RuntimeError("At the observatories jaeger must run on sdss5-fps.")

    if verbose > 0 and quiet:
        raise click.UsageError("--quiet and --verbose are mutually exclusive.")

    if config_file:
        config.load(config_file)

    actor_config = config.get("actor", {})

    if verbose == 1:
        log.sh.setLevel(logging.INFO)
        actor_config["verbose"] = logging.INFO
    elif verbose == 2:
        log.sh.setLevel(logging.DEBUG)
        actor_config["verbose"] = logging.DEBUG
    elif verbose >= 3:
        log.sh.setLevel(logging.DEBUG)
        can_log.sh.setLevel(logging.DEBUG)
        actor_config["verbose"] = logging.DEBUG

    if quiet:
        log.handlers.remove(log.sh)
        warnings.simplefilter("ignore")
        actor_config["verbose"] = 100

    if sextant:
        sextant_file = os.path.join(os.path.dirname(__file__), "etc/sextant.yaml")
        config["files"]["ieb_config"] = sextant_file
        log.debug(f"Using internal IEB sextant onfiguration file {sextant_file}.")

    if sextant or "sextants/" in config["files"]["ieb_config"]:
        enable_low_temperature = False
    else:
        enable_low_temperature = True

    if virtual is True:
        profile = "virtual"

    ctx.obj = FPSWrapper(
        profile,
        ieb=ieb,
        npositioners=npositioners,
        enable_low_temperature=enable_low_temperature,
        use_lock=not no_lock,
    )


LOG_FILE = os.path.join(
    os.environ.get("ACTOR_DAEMON_LOG_DIR", "$HOME/logs"),
    "jaeger/jaeger.log",
)


@jaeger.group(
    cls=DaemonGroup,
    prog="jaeger_actor",
    workdir=os.getcwd(),
    log_file=LOG_FILE,
)
@click.option(
    "--no-tron",
    is_flag=True,
    help="Does not connect to Tron.",
)
@pass_fps
@cli_coro
async def actor(fps_maker, no_tron):
    """Runs the actor."""

    try:
        from jaeger.actor import JaegerActor
    except ImportError:
        raise ImportError("CLU needs to be installed to run jaeger as an actor.")

    actor_config = config["actor"].copy()
    actor_config.pop("status", None)

    if no_tron:
        actor_config.pop("tron", None)

    async with fps_maker as fps:
        actor_: Any = await JaegerActor.from_config(actor_config, fps).start()
        await actor_.start_status_server(
            config["actor"]["status"]["port"],
            delay=config["actor"]["status"]["delay"],
        )
        await actor_.run_forever()


@jaeger.command(name="upgrade-firmware")
@click.argument(
    "firmware-file",
    nargs=1,
    type=click.Path(exists=True),
)
@click.argument(
    "SEXTANTS",
    type=int,
    nargs=-1,
    required=False,
)
@click.option(
    "-y",
    "--yes",
    is_flag=True,
    help="Do not ask for confirmation.",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Forces skipping of invalid positioners",
)
@click.option(
    "-p",
    "--positioners",
    type=str,
    help="Comma-separated positioners to upgrade",
)
@click.option(
    "-n",
    "--no-cycle",
    is_flag=True,
    help="Does not power cycle positioners before upgrading each sextant",
)
@click.option(
    "-o",
    "--all-on",
    is_flag=True,
    help="Powers on all sextants after a successful upgrade.",
)
@pass_fps
@cli_coro
async def upgrade_firmware(
    fps_maker,
    firmware_file,
    force,
    positioners,
    no_cycle,
    sextants,
    all_on,
    yes,
):
    """Upgrades the firmaware."""

    if positioners is not None:
        positioners = [int(positioner.strip()) for positioner in positioners.split(",")]

    fps_maker.initialise = False

    sextants = sextants or (1, 2, 3, 4, 5, 6)

    if not yes:
        click.confirm(
            f"Upgrade firmware for sextant(s) {', '.join(map(str, sextants))}?",
            default=False,
            abort=True,
        )

    async with fps_maker as fps:

        ps_devs = []
        if fps.ieb and no_cycle is False:
            for module in fps.ieb.modules.values():
                for dev_name in module.devices:
                    if dev_name.upper().startswith("PS"):
                        ps_devs.append(dev_name)

        for sextant in sextants:
            log.info(f"Upgrading firmware on sextant {sextant}.")

            if fps.ieb and no_cycle is False:
                log.info("Power cycling positioners")

                for dev in ps_devs:
                    await fps.ieb.get_device(dev).open()
                    await asyncio.sleep(1)

                await asyncio.sleep(5)

                dev = "PS" + str(sextant)
                await fps.ieb.get_device(dev).close()

                await asyncio.sleep(3)

            await fps.initialise(start_pollers=False)

            await load_firmware(
                fps,
                firmware_file,
                positioners=positioners,
                force=force,
                show_progressbar=True,
            )

    if all_on is True:
        log.info("Powering on sextants.")
        for sextant in sextants:
            await fps.ieb.get_device(f"PS{sextant}").close()
            await asyncio.sleep(3)


@jaeger.command()
@click.argument("positioner-id", nargs=1, type=int)
@click.option(
    "--motors/--no-motors",
    is_flag=True,
    default=True,
    help="Run the motor calibration.",
)
@click.option(
    "--datums/--no-datums",
    is_flag=True,
    default=True,
    help="Run the datum calibration.",
)
@click.option(
    "--cogging/--no-cogging",
    is_flag=True,
    default=True,
    help="Run the cogging calibration (can take a long time).",
)
@pass_fps
@cli_coro
async def calibrate(fps_maker, positioner_id, motors, datums, cogging):
    """Runs a full calibration on a positioner."""

    fps_maker.initialise = False
    fps_maker.danger = True

    async with fps_maker as fps:
        await fps.initialise(start_pollers=False)
        await calibrate_positioner(
            fps,
            positioner_id,
            motors=motors,
            datums=datums,
            cogging=cogging,
        )


@jaeger.command()
@click.argument("positioner_id", metavar="POSITIONER", type=int)
@click.argument("alpha", metavar="ALPHA", type=float)
@click.argument("beta", metavar="BETA", type=float)
@click.option(
    "--speed",
    type=(float, float),
    default=(None, None),
    help="The speed for the alpha and beta motors.",
    show_default=True,
)
@pass_fps
@cli_coro
async def goto(fps_maker, positioner_id, alpha, beta, speed=None):
    """Moves a robot to a given position."""

    if alpha < 0 or alpha >= 360:
        raise click.UsageError("alpha must be in the range [0, 360)")

    if beta < 0 or beta >= 360:
        raise click.UsageError("beta must be in the range [0, 360)")

    if speed[0] or speed[1]:
        if speed[0] < 0 or speed[0] >= 3000 or speed[1] < 0 or speed[1] >= 3000:
            raise click.UsageError("speed must be in the range [0, 3000)")

    async with fps_maker as fps:

        positioner = fps.positioners[positioner_id]
        result = await positioner.initialise()
        if not result:
            log.error("positioner is not connected or failed to initialise.")
            return

        await positioner.goto(alpha=alpha, beta=beta, speed=(speed[0], speed[1]))

    return


@jaeger.command(name="set-positions")
@click.argument("positioner_id", metavar="POSITIONER", type=int)
@click.argument("alpha", metavar="ALPHA", type=float)
@click.argument("beta", metavar="BETA", type=float)
@pass_fps
@cli_coro
async def set_positions(fps_maker, positioner_id, alpha, beta):
    """Sets the position of the alpha and beta arms."""

    if alpha < 0 or alpha >= 360:
        raise click.UsageError("alpha must be in the range [0, 360)")

    if beta < 0 or beta >= 360:
        raise click.UsageError("beta must be in the range [0, 360)")

    async with fps_maker as fps:

        positioner = fps.positioners[positioner_id]

        result = await positioner.set_position(alpha, beta)

        if not result:
            log.error("failed to set positions.")
            return

        log.info(f"positioner {positioner_id} set to {(alpha, beta)}.")


@jaeger.command()
@click.argument("positioner_id", metavar="POSITIONER", type=int)
@click.option(
    "-n",
    "--moves",
    type=int,
    help="Number of moves to perform. Otherwise runs forever.",
)
@click.option(
    "--alpha",
    type=(int, int),
    default=(0, 360),
    help="Range of alpha positions.",
    show_default=True,
)
@click.option(
    "--beta",
    type=(int, int),
    default=(0, 180),
    help="Range of beta positions.",
    show_default=True,
)
@click.option(
    "--speed",
    type=(int, int),
    default=(500, 1500),
    help="Range of speed.",
    show_default=True,
)
@click.option(
    "-f",
    "--skip-errors",
    is_flag=True,
    help="If an error occurs, ignores it and commands another move.",
)
@pass_fps
@cli_coro
async def demo(
    fps_maker,
    positioner_id,
    alpha=None,
    beta=None,
    speed=None,
    moves=None,
    skip_errors=False,
):
    """Moves a robot to random positions."""

    if (alpha[0] >= alpha[1]) or (alpha[0] < 0 or alpha[1] > 360):
        raise click.UsageError("alpha must be in the range [0, 360)")

    if (beta[0] >= beta[1]) or (beta[0] < 0 or beta[1] > 360):
        raise click.UsageError("beta must be in the range [0, 360)")

    if (speed[0] >= speed[1]) or (speed[0] < 0 or speed[1] >= 3000):
        raise click.UsageError("speed must be in the range [0, 3000)")

    async with fps_maker as fps:

        positioner = fps.positioners[positioner_id]
        result = await positioner.initialise()
        if not result:
            log.error("positioner is not connected or failed to initialise.")
            return

        done_moves = 0
        while True:

            alpha_move = numpy.random.randint(low=alpha[0], high=alpha[1])
            beta_move = numpy.random.randint(low=beta[0], high=beta[1])
            alpha_speed = numpy.random.randint(low=speed[0], high=speed[1])
            beta_speed = numpy.random.randint(low=speed[0], high=speed[1])

            warnings.warn(f"running step {done_moves+1}")

            result = await positioner.goto(
                alpha=alpha_move, beta=beta_move, speed=(alpha_speed, beta_speed)
            )

            if result is False:
                if skip_errors is False:
                    return
                else:
                    warnings.warn(
                        "an error happened but ignoring it because skip-error=True"
                    )
                    continue

            done_moves += 1

            if moves is not None and done_moves == moves:
                return


@jaeger.command()
@click.argument("positioner_id", metavar="POSITIONER", type=int, required=False)
@pass_fps
@cli_coro
async def home(fps_maker, positioner_id):
    """Initialise datums."""

    async with fps_maker as fps:

        if positioner_id is None:
            positioners = fps.positioners.values()
        else:
            positioners = [fps.positioners[positioner_id]]

        valid_positioners = [
            positioner for positioner in positioners if positioner.status.initialised
        ]

        await asyncio.gather(*[positioner.home() for positioner in valid_positioners])

    return


@jaeger.command()
@click.argument("positioner_id", metavar="POSITIONER", type=int)
@pass_fps
@cli_coro
async def status(fps_maker: FPSWrapper, positioner_id: int):
    """Returns the status of a positioner with low-level initialisation."""

    fps_maker.initialise = False

    async with fps_maker as fps:

        pos = Positioner(positioner_id, fps)

        try:
            await pos.update_firmware_version()
            print(f"Firmware: {pos.firmware}")
            print(f"Bootloader: {pos.is_bootloader()}")
        except Exception as err:
            raise JaegerError(f"Failed retrieving firmware: {err}")

        try:
            await pos.update_status()
            bit_names = ", ".join(bit.name for bit in pos.status.active_bits)
            print(f"Status: {pos.status.value} ({bit_names})")
        except Exception as err:
            raise JaegerError(f"Failed retrieving status: {err}")

        try:
            await pos.update_position()
            print(f"Position: {pos.position}")
        except Exception as err:
            raise JaegerError(f"Failed retrieving position: {err}")


@jaeger.command()
@click.option(
    "--collision-buffer",
    type=click.FloatRange(1.6, 3.0),
    help="Custom collision buffer",
)
@click.option(
    "--force",
    is_flag=True,
    help="Execute unwind even in presence of deadlocks.",
)
@pass_fps
@cli_coro
async def unwind(fps_maker, collision_buffer: float | None = None, force: bool = False):
    """Unwinds the array."""

    from jaeger.kaiju import unwind

    async with fps_maker as fps:

        await fps.update_position()
        positions = {p.positioner_id: (p.alpha, p.beta) for p in fps.values()}

        log.info("Calculating trajectory.")
        trajectory = await unwind(
            positions,
            collision_buffer=collision_buffer,
            force=force,
        )

        log.info("Executing unwind trajectory.")
        await fps.send_trajectory(trajectory)

    return


@jaeger.command()
@click.argument("EXPLODE-DEG", type=float)
@pass_fps
@cli_coro
async def explode(fps_maker, explode_deg: float):
    """Explodes the array."""

    from jaeger.kaiju import explode

    async with fps_maker as fps:

        await fps.update_position()
        positions = {p.positioner_id: (p.alpha, p.beta) for p in fps.values()}

        log.info("Calculating trajectory.")
        trajectory = await explode(positions, explode_deg=explode_deg)

        log.info("Executing explode trajectory.")
        await fps.send_trajectory(trajectory)

    return


@jaeger.command()
@click.argument("PATH", required=False, type=click.Path(exists=False, dir_okay=False))
@click.option("--collision-buffer", type=float, help="The collision buffer.")
@pass_fps
@cli_coro
async def snapshot(
    fps_maker: FPSWrapper,
    path: Optional[str] = None,
    collision_buffer: float | None = None,
):
    """Takes a snapshot image."""

    if path is not None:
        path = str(path)

    async with fps_maker as fps:

        await fps.update_position()
        filename = await fps.save_snapshot(path, collision_buffer=collision_buffer)

    print(f"Snapshot saved to {filename}")


@jaeger.command()
@pass_fps
@cli_coro
async def unlock(fps_maker: FPSWrapper):
    """Unlocks the FPS."""

    warnings.filterwarnings(
        "ignore",
        message=".+FPS was collided and has been locked.+",
        category=JaegerUserWarning,
    )

    async with fps_maker as fps:
        await fps.unlock()

    if os.path.exists(LOCK_FILE):
        os.remove(LOCK_FILE)


if __name__ == "__main__":
    jaeger()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-13
# @Filename: commands.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import pathlib

from typing import TYPE_CHECKING, List, Tuple

import click
import numpy

import clu

from jaeger import config
from jaeger.commands import CommandID, SetCurrent, Trajectory
from jaeger.exceptions import JaegerError, TrajectoryError
from jaeger.utils import get_goto_move_time

from . import jaeger_parser


if TYPE_CHECKING:
    from clu import Command

    from jaeger import FPS
    from jaeger.actor import JaegerActor


__all__ = [
    "goto",
    "speed",
    "initialise",
    "stop",
    "hall",
    "trajectory",
    "unlock",
    "current",
    "status",
]


def check_positioners(positioner_ids, command, fps, initialised=False):
    """Checks if some of the positioners are not connected."""

    if any([pid not in fps.positioners for pid in positioner_ids]):
        command.fail(error="some positioners are not connected.")
        return False

    if initialised:
        if any([not fps[pid].initialised for pid in positioner_ids]):
            command.fail(error="some positioners are not initialised.")
            return False

    return True


@jaeger_parser.command()
@click.argument("POSITIONER-IDS", type=int, nargs=-1)
@click.argument("ALPHA", type=click.FloatRange(-360.0, 360.0))
@click.argument("BETA", type=click.FloatRange(-360.0, 360.0))
@click.option(
    "-r",
    "--relative",
    is_flag=True,
    help="Whether this is a relative move",
)
@click.option(
    "-s",
    "--speed",
    type=click.FloatRange(100.0, 4000.0),
    nargs=2,
    help="The speed of both alpha and beta arms, in RPS on the input.",
)
@click.option(
    "-a",
    "--all",
    is_flag=True,
    default=False,
    help="Applies to all valid positioners.",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Forces a move to happen.",
)
@click.option(
    "--use-sync/-no-use-sync",
    " /-S",
    default=True,
    help="Whether to use the SYNC line to start the trajectory.",
)
async def goto(
    command: Command[JaegerActor],
    fps: FPS,
    positioner_ids: Tuple[int, ...],
    alpha: float,
    beta: float,
    speed: Tuple[float, float],
    all: bool = False,
    force: bool = False,
    relative: bool = False,
    use_sync: bool = True,
):
    """Sends positioners to a given (alpha, beta) position."""

    assert command.actor

    if all:
        if not force:
            return command.fail(error="Use --force to move all positioners at once.")
        pids: List[int] = list(fps.positioners.keys())
    else:
        pids = list(positioner_ids)

    if not relative:
        if alpha < 0 or beta < 0:
            return command.fail(error="Negative angles only allowed in relative mode.")

    if not check_positioners(pids, command, fps, initialised=True):
        return

    if fps.moving:
        return command.fail(error="FPS is moving. Cannot send goto.")

    if speed is None:
        default_speed = config["positioner"]["motor_speed"]
        speed = (default_speed, default_speed)

    max_time = 0.0

    for pid in pids:

        # Manually calculate the max move time we'll encounter.
        p_alpha, p_beta = fps[pid].position

        if p_alpha is None or p_beta is None:
            return command.fail(error="Some positioners do not know their positions.")

        delta_alpha = abs(p_alpha - alpha) if not relative else alpha
        delta_beta = abs(p_beta - beta) if not relative else beta

        time_alpha = get_goto_move_time(
            delta_alpha,
            speed=speed[0] or fps[pid].speed[0],
        )
        time_beta = get_goto_move_time(
            delta_beta,
            speed=speed[1] or fps[pid].speed[1],
        )

        if time_alpha > max_time:
            max_time = time_alpha
        if time_beta > max_time:
            max_time = time_beta

    command.info(move_time=round(max_time, 2))

    try:
        await command.actor.fps.goto(
            pids,
            alpha,
            beta,
            speed=speed,
            relative=relative,
            use_sync_line=use_sync,
        )
    except JaegerError as err:
        return command.fail(error=f"Goto command failed: {err}")

    command.finish(text="Destination reached")


@jaeger_parser.command()
@click.argument("POSITIONER-ID", type=int, nargs=-1)
@click.argument("ALPHA", type=click.FloatRange(50.0, 5000.0))
@click.argument("BETA", type=click.FloatRange(50.0, 5000.0))
@click.option(
    "-a",
    "--all",
    is_flag=True,
    default=False,
    help="applies to all valid positioners.",
)
async def speed(command, fps, positioner_id, alpha, beta, all):
    """Sets the ``(alpha, beta)`` speed in RPM on the input."""

    if all:
        positioner_id = list(fps.positioners.keys())

    if not check_positioners(positioner_id, command, fps, initialised=True):
        return

    if fps.moving:
        return command.fail(error="FPS is moving. Cannot send set_speed.")

    tasks = []
    for pid in positioner_id:
        tasks.append(fps.positioners[pid].set_speed(alpha, beta))

    result = await clu.as_complete_failer(tasks, on_fail_callback=fps.stop_trajectory)

    if not result[0]:
        error_message = result[1] or "set speed command failed"
        command.set_status(clu.CommandStatus.FAILED, error=error_message)
    else:
        command.set_status(clu.CommandStatus.DONE, text="Set speed done")


@jaeger_parser.command()
@click.argument("POSITIONER-ID", type=int, nargs=-1)
@click.option(
    "--datums",
    is_flag=True,
    help="If set, initialises the datums.",
)
async def initialise(command, fps, positioner_id, datums=False):
    """Initialises positioners."""

    if not check_positioners(positioner_id, command, fps):
        return

    tasks = []
    for pid in positioner_id:
        tasks.append(fps.positioners[pid].initialise(initialise_datums=datums))

    result = await clu.as_complete_failer(tasks)

    if not result[0]:
        error_message = result[1] or "initialise failed"
        command.set_status(clu.CommandStatus.FAILED, error=error_message)
    else:
        command.set_status(clu.CommandStatus.DONE, text="Initialisation complete")


@jaeger_parser.command()
@click.argument("POSITIONER-IDS", type=int, nargs=-1, required=False)
@click.option(
    "-f",
    "--full",
    is_flag=True,
    default=False,
    help="outputs more statuses.",
)
async def status(command, fps, positioner_ids, full):
    """Reports the position and status bit of a list of positioners."""

    positioner_ids = positioner_ids or list(fps.positioners.keys())

    if not check_positioners(positioner_ids, command, fps):
        return

    try:
        await fps.update_status(positioner_ids=0)
        await fps.update_position(positioner_ids=positioner_ids)
    except JaegerError as err:
        return command.fail(error=f"Failed reporting status: {err}")

    n_trajs = (
        await fps.send_command(
            CommandID.GET_NUMBER_TRAJECTORIES,
            positioner_ids=positioner_ids,
        )
    ).get_replies()

    command.info(locked=fps.locked)

    command.info(n_positioners=len(fps.positioners))

    for pid in sorted(positioner_ids):
        p = fps[pid]

        alpha_pos = -999 if p.alpha is None else numpy.round(p.alpha, 4)
        beta_pos = -999 if p.beta is None else numpy.round(p.beta, 4)

        n_trajs_pid = n_trajs[pid] if n_trajs[pid] is not None else "?"

        if pid in fps.positioner_to_bus:
            interface, bus = fps.positioner_to_bus[pid]
            interface = fps.can.interfaces.index(interface) + 1
        else:
            interface = -1
            bus = -1

        command.write(
            "i",
            positioner_status=[
                p.positioner_id,
                alpha_pos,
                beta_pos,
                f"0x{int(p.status):x}",
                p.initialised,
                p.is_bootloader() or False,
                p.firmware or "?",
                interface,
                bus,
                n_trajs_pid,
            ],
        )

    command.info(fps_status=f"0x{fps.status.value:x}")

    if full:
        await clu.Command("ieb status", parent=command).parse()

    command.set_status(clu.CommandStatus.DONE)


@jaeger_parser.command()
@click.argument("POSITIONER-ID", type=int, nargs=-1, required=False)
@click.argument("ALPHA", type=click.FloatRange(0.0, 100.0))
@click.argument("BETA", type=click.FloatRange(0.0, 100.0))
@click.option(
    "-a",
    "--all",
    is_flag=True,
    default=False,
    help="applies to all connected positioners.",
)
async def current(command, fps, positioner_id, alpha, beta, all):
    """Sets the current of the positioner."""

    if all:
        positioner_id = [pid for pid in fps.positioners if fps[pid].initialised]

    if len(positioner_id) == 0:
        return command.fail(error="no positioners provided.")

    if not check_positioners(positioner_id, command, fps):
        return

    if fps.moving:
        return command.fail(error="FPS is moving. Cannot send set current.")

    commands = [
        fps.send_command(SetCurrent(pid, alpha=alpha, beta=beta))
        for pid in positioner_id
    ]
    await asyncio.gather(*commands)

    return command.finish(text="current changed.")


@jaeger_parser.command()
async def stop(command, fps):
    """Stops the positioners and clear flags."""

    await fps.stop_trajectory()
    await fps.update_status(timeout=0.1)
    await fps.update_position()

    command.set_status(clu.CommandStatus.DONE, text="Trajectory aborted")


@jaeger_parser.command()
async def unlock(command, fps):
    """Unlocks the FPS."""

    result = await fps.unlock()

    if result:
        command.info(locked=False)
        return command.finish(text="FPS unlocked")
    else:
        return command.fail(error="failed to unlock FPS")


@jaeger_parser.command()
@click.argument("PATH", type=str)
async def trajectory(command, fps, path):
    """Sends a trajectory from a file."""

    if fps.moving:
        return command.fail(error="FPS is moving. Cannot send trajectory.")

    if fps.locked:
        return command.fail(
            error=f"FPS is locked by {fps.locked_by}. Cannot send trajectory."
        )

    path = pathlib.Path(path).expanduser()
    if not path.exists():
        raise click.BadParameter(f"path {path!s} does not exist.")

    try:

        trajectory = Trajectory(fps, path)

        command.debug(text="sending trajectory ...")
        await trajectory.send()
        if trajectory.failed:
            return command.fail(error="failed sending trajectory with unknown error.")

        command.debug(text=f"trajectory sent in {trajectory.data_send_time:.2f} s.")
        command.info(
            text=f"move will take {trajectory.move_time:.2f} s",
            move_time=f"{trajectory.move_time:.2f}",
        )

        await trajectory.start()
        if trajectory.failed:
            return command.fail(error="failed starting trajectory with unknown error.")

        return command.finish(text="trajectory completed.")

    except TrajectoryError as ee:
        return command.fail(error=str(ee))


@jaeger_parser.group()
def hall():
    """Turns the hall sensor on/off."""

    pass


@hall.command()
@click.argument("POSITIONER-ID", type=int, nargs=-1, required=False)
async def on(command, fps, positioner_id):
    """Turns the hall sensor on."""

    if positioner_id is None:
        positioner_id = list(fps.positioners.keys())

    if not check_positioners(positioner_id, command, fps, initialised=False):
        return

    command.debug("Turning hall sensors ON")
    await fps.send_command("HALL_ON", positioner_ids=positioner_id)

    command.debug("Waiting 5 seconds ...")
    await asyncio.sleep(5)

    command.finish()


@hall.command()
@click.argument("POSITIONER-ID", type=int, nargs=-1, required=False)
async def off(command, fps, positioner_id):
    """Turns the hall sensor off."""

    if positioner_id is None:
        positioner_id = list(fps.positioners.keys())

    if not check_positioners(positioner_id, command, fps, initialised=False):
        return

    command.debug("Turning hall sensors OFF")
    await fps.send_command("HALL_OFF", positioner_ids=positioner_id)

    command.finish()


@jaeger_parser.group()
def led():
    """Turns the positioner LED on/off."""

    pass


@led.command(name="on")
@click.argument("POSITIONER-ID", type=int, nargs=-1, required=False)
async def led_on(command, fps, positioner_id):
    """Turns the LED on."""

    if positioner_id is None:
        positioner_id = list(fps.positioners.keys())

    if not check_positioners(positioner_id, command, fps, initialised=False):
        return

    command.debug("Turning LED on")
    await fps.send_command("SWITCH_LED_ON", positioner_ids=positioner_id)

    command.finish()


@led.command(name="off")
@click.argument("POSITIONER-ID", type=int, nargs=-1, required=False)
async def led_off(command, fps, positioner_id):
    """Turns the LED off."""

    if positioner_id is None:
        positioner_id = list(fps.positioners.keys())

    if not check_positioners(positioner_id, command, fps, initialised=False):
        return

    command.debug("Turning LED off")
    await fps.send_command("SWITCH_LED_OFF", positioner_ids=positioner_id)

    command.finish()


@jaeger_parser.command()
async def reload(command, fps):
    """Reinitialise the FPS."""

    try:
        await fps.initialise(start_pollers=fps.pollers.running)
    except BaseException as err:
        return command.fail(error=f"Initialisation failed: {err}")

    return command.finish(text="FPS was reinitialised.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-10-03
# @Filename: goto.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

import numpy

import jaeger
from jaeger import config
from jaeger.commands import Command, CommandID
from jaeger.exceptions import JaegerError
from jaeger.utils import (
    bytes_to_int,
    get_goto_move_time,
    int_to_bytes,
    motor_steps_to_angle,
)

from .trajectory import send_trajectory


if TYPE_CHECKING:
    from jaeger.fps import FPS


__all__ = [
    "GoToDatums",
    "GoToDatumAlpha",
    "GoToDatumBeta",
    "GotoAbsolutePosition",
    "GotoRelativePosition",
    "SetActualPosition",
    "SetSpeed",
    "SetCurrent",
    "goto",
]


TIME_STEP = jaeger.config["positioner"]["time_step"]


class GoToDatums(Command):
    """Initialises and zeroes the positioner."""

    command_id = CommandID.GO_TO_DATUMS
    broadcastable = False
    move_command = True


class GoToDatumAlpha(Command):
    """Initialises and zeroes the alpha arm of the positioner."""

    command_id = CommandID.GO_TO_DATUM_ALPHA
    broadcastable = False
    move_command = True


class GoToDatumBeta(Command):
    """Initialises and zeroes the beta arm of the positioner."""

    command_id = CommandID.GO_TO_DATUM_BETA
    broadcastable = False
    move_command = True


class GotoAbsolutePosition(Command):
    """Moves alpha and beta to absolute positions in degrees."""

    command_id = CommandID.GO_TO_ABSOLUTE_POSITION
    broadcastable = False
    move_command = True

    def __init__(
        self,
        positioner_ids: int | List[int],
        alpha: float | None = None,
        beta: float | None = None,
        **kwargs,
    ):

        if alpha is not None and beta is not None:
            alpha_steps, beta_steps = motor_steps_to_angle(alpha, beta, inverse=True)

            alpha_bytes = int_to_bytes(alpha_steps, dtype="i4")
            beta_bytes = int_to_bytes(beta_steps, dtype="i4")

            data = alpha_bytes + beta_bytes
            kwargs["data"] = data

        super().__init__(positioner_ids, **kwargs)

    @staticmethod
    def decode(data):
        """Decodes message data into alpha and beta moves."""

        alpha_steps = bytes_to_int(data[0:4], dtype="i4")
        beta_steps = bytes_to_int(data[4:8], dtype="i4")

        return motor_steps_to_angle(alpha_steps, beta_steps)

    def get_replies(self) -> Dict[int, Tuple[float, float]]:
        return self.get_move_time()

    def get_move_time(self):
        """Returns the time needed to move to the commanded position.

        Raises
        ------
        ValueError
            If no reply has been received or the data cannot be parsed.

        """

        move_times = {}
        for reply in self.replies:
            data = reply.data

            alpha = bytes_to_int(data[0:4], dtype="i4")
            beta = bytes_to_int(data[4:], dtype="i4")

            move_times[reply.positioner_id] = [alpha * TIME_STEP, beta * TIME_STEP]

        return move_times


class GotoRelativePosition(GotoAbsolutePosition):
    """Moves alpha and beta a relative number of degrees."""

    command_id = CommandID.GO_TO_RELATIVE_POSITION
    broadcastable = False
    move_command = True


class SetActualPosition(Command):
    """Sets the current position of the alpha and beta arms."""

    command_id = CommandID.SET_ACTUAL_POSITION
    broadcastable = False
    safe = True
    move_command = True  # Technically not a move command but we don't
    # want to issue it during a move.

    def __init__(
        self,
        positioner_ids: int | List[int],
        alpha: float | None = None,
        beta: float | None = None,
        **kwargs,
    ):

        if alpha is not None and beta is not None:

            alpha_steps, beta_steps = motor_steps_to_angle(alpha, beta, inverse=True)

            alpha_bytes = int_to_bytes(int(alpha_steps), dtype="i4")
            beta_bytes = int_to_bytes(int(beta_steps), dtype="i4")

            data = alpha_bytes + beta_bytes

            kwargs["data"] = data

        super().__init__(positioner_ids, **kwargs)


class SetSpeed(Command):
    """Sets the speeds of the alpha and beta motors."""

    command_id = CommandID.SET_SPEED
    broadcastable = False
    safe = True
    move_command = False

    def __init__(
        self,
        positioner_ids: int | List[int],
        alpha: float | None = None,
        beta: float | None = None,
        **kwargs,
    ):

        if alpha is not None and beta is not None:
            assert alpha >= 0 and beta >= 0, "invalid speed."

            data = int_to_bytes(int(alpha)) + int_to_bytes(int(beta))
            kwargs["data"] = data

        super().__init__(positioner_ids, **kwargs)

    @staticmethod
    def encode(alpha, beta):
        """Encodes the alpha and beta speed as bytes."""

        data_speed = int_to_bytes(int(alpha)) + int_to_bytes(int(beta))

        return data_speed


class SetCurrent(Command):
    """Sets the current of the alpha and beta motors."""

    command_id = CommandID.SET_CURRENT
    broadcastable = False
    safe = True
    move_command = True

    def __init__(
        self,
        positioner_ids: int | List[int],
        alpha: float | None = None,
        beta: float | None = None,
        **kwargs,
    ):

        if alpha is not None and beta is not None:
            assert alpha >= 0 and beta >= 0, "invalid current."

            data = int_to_bytes(int(alpha)) + int_to_bytes(int(beta))
            kwargs["data"] = data

        super().__init__(positioner_ids, **kwargs)


async def goto(
    fps: FPS,
    positioner_ids: List[int],
    alpha: float | list | numpy.ndarray,
    beta: float | list | numpy.ndarray,
    speed: Optional[Tuple[float, float]] = None,
    relative: bool = False,
    use_sync_line: bool = None,
):
    """Send positioners to a given position using a trajectory.

    Parameters
    ----------
    fps
        The `.FPS` instance.
    positioner_ids
        The list of positioner_ids to command.
    alpha
        The alpha angle. Can be an array with the same size of the list of positioner
        IDs. Otherwise sends all the positioners to the same angle.
    beta
        The beta angle.
    speed
        As a tuple, the alpha and beta speeds to use. If `None`, uses the default ones.
    relative
        If `True`, ``alpha`` and ``beta`` are considered relative angles.
    use_sync_line
        Whether to use the SYNC line to start the trajectories.

    """

    if not isinstance(alpha, (list, tuple, numpy.ndarray)):
        alpha = numpy.tile(alpha, len(positioner_ids))
    if not isinstance(beta, (list, tuple, numpy.ndarray)):
        beta = numpy.tile(beta, len(positioner_ids))

    alpha = numpy.array(alpha)
    beta = numpy.array(beta)

    assert len(alpha) == len(positioner_ids) and len(beta) == len(positioner_ids)

    if alpha is None or beta is None:
        raise JaegerError("alpha and beta must be non-null.")

    if speed is None:
        default_speed = config["positioner"]["motor_speed"]
        speed = (default_speed, default_speed)
    else:
        if len(speed) != 2 or speed[0] is None or speed[1] is None:
            raise JaegerError("Invalid speed.")

    speed_array = numpy.array(speed)
    if numpy.any(speed_array <= 0) or numpy.any(speed_array > 5000):
        raise JaegerError("Speed out of bounds.")

    await fps.update_position(positioner_ids=positioner_ids)

    trajectories = {}
    for i, pid in enumerate(positioner_ids):
        pos = fps[pid]

        if pos.alpha is None or pos.beta is None:
            raise JaegerError(f"Positioner {pid}: cannot goto with unknown position.")

        current_alpha = pos.alpha
        current_beta = pos.beta

        if relative is True:
            alpha_end = current_alpha + alpha[i]
            beta_end = current_beta + beta[i]
        else:
            alpha_end = alpha[i]
            beta_end = beta[i]

        alpha_delta = abs(alpha_end - current_alpha)
        beta_delta = abs(beta_end - current_beta)

        time_end = [
            get_goto_move_time(alpha_delta, speed=speed[0]),
            get_goto_move_time(beta_delta, speed=speed[1]),
        ]

        trajectories[pid] = {
            "alpha": [(current_alpha, 0.1), (alpha_end, time_end[0] + 0.1)],
            "beta": [(current_beta, 0.1), (beta_end, time_end[1] + 0.1)],
        }

    return await send_trajectory(fps, trajectories, use_sync_line=use_sync_line)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-04-24
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json
import logging
import os
from tempfile import NamedTemporaryFile

import clu
import clu.protocol
from clu.tools import ActorHandler

import jaeger
from jaeger import FPS, __version__, log
from jaeger.exceptions import JaegerUserWarning


__all__ = ["JaegerActor"]


def merge_json(base: str, custom: str | None, write_temporary_file=False):
    """Merges two JSON files. Optional writes the result as a temporary file."""

    if custom is None:
        return base

    if os.path.samefile(base, custom):
        return base
    else:
        base_json = json.loads(open(base, "r").read())
        custom_json = json.loads(open(custom, "r").read())
        base_json["properties"].update(custom_json["properties"])

        if write_temporary_file is False:
            return base_json
        else:
            tmpfile = NamedTemporaryFile(delete=False, mode="w+")
            json.dump(base_json, tmpfile)
            tmpfile.flush()
            return tmpfile.name


class JaegerActor(clu.LegacyActor):
    """The jaeger SDSS-style actor."""

    def __init__(self, fps: FPS, *args, ieb_status_delay=60, **kwargs):

        jaeger.actor_instance = self

        self.fps = fps

        # This is mostly for the miniwoks. If the schema file is not the base
        # one, merge them.
        base = os.path.join(os.path.dirname(__file__), "..")

        base_schema = os.path.realpath(os.path.join(base, "etc/schema.json"))

        schema = kwargs.get("schema", None)
        c_schema = os.path.realpath(os.path.join(base, schema)) if schema else None

        kwargs["schema"] = merge_json(base_schema, c_schema, write_temporary_file=True)

        # Pass the FPS instance as the second argument to each parser
        # command (the first argument is always the actor command).
        self.parser_args = [fps]

        super().__init__(*args, **kwargs)

        self.version = __version__

        # Add ActorHandler to log and to the warnings logger.
        self.actor_handler = ActorHandler(
            self,
            level=logging.WARNING,
            filter_warnings=[JaegerUserWarning],
        )
        log.addHandler(self.actor_handler)
        if log.warnings_logger:
            log.warnings_logger.addHandler(self.actor_handler)

        self.__status_watcher_task = asyncio.create_task(self._status_watcher())

    async def start_status_server(self, port, delay=1):
        """Starts a server that outputs the status as a JSON on a timer."""

        self.status_server = clu.protocol.TCPStreamPeriodicServer(
            self.host,
            port,
            periodic_callback=self._report_status_cb,
            sleep_time=delay,
        )

        await self.status_server.start()

        self.log.info(f"starting status server on {self.host}:{port}")

    async def _report_status_cb(self, transport):
        """Reports the status to the status server."""

        status = await self.fps.report_status()
        status_json = json.dumps(status)

        transport.write(status_json.encode() + "\n".encode())

        return status

    async def _status_watcher(self):
        """Listens to the status async generator."""

        async for status in self.fps.async_status():
            self.write("i", fps_status=f"0x{status.value:x}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-11-01
# @Filename: fvc.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import logging
import os
import pathlib
import warnings

from typing import TYPE_CHECKING, Optional

import numpy
import pandas
import sep
from astropy.io import fits
from astropy.table import Table
from matplotlib import pyplot as plt

from clu.command import Command
from clu.legacy.tron import TronConnection
from coordio.defaults import calibration
from coordio.transforms import RoughTransform, ZhaoBurgeTransform

from jaeger import config, log
from jaeger.exceptions import FVCError, JaegerUserWarning, TrajectoryError
from jaeger.fps import FPS
from jaeger.ieb import IEB
from jaeger.kaiju import get_path_pair_in_executor, get_robot_grid
from jaeger.target.tools import wok_to_positioner
from jaeger.utils import run_in_executor


if TYPE_CHECKING:
    from jaeger.actor import JaegerActor

__all__ = ["FVC"]


class FVC:
    """Focal View Camera class."""

    fibre_data: Optional[pandas.DataFrame]
    centroids: Optional[pandas.DataFrame]
    offsets: Optional[pandas.DataFrame]

    image_path: Optional[str]
    raw_hdu: Optional[fits.ImageHDU]
    proc_hdu: Optional[fits.ImageHDU]

    fitrms: float
    k: float

    def __init__(self, site: str, command: Optional[Command[JaegerActor]] = None):

        if len(calibration.positionerTable) == 0:
            warnings.warn(
                "FPS calibrations not loaded or the array is empty.",
                JaegerUserWarning,
            )

        self.site = site

        self.command = command
        self.fps = FPS.get_instance()

        self.reset()

    def reset(self):
        """Resets the instance."""

        self.fibre_data = None
        self.centroids = None
        self.offsets = None

        self.image_path = None
        self.raw_hdu = None
        self.proc_hdu = None

        self.k = 1
        self.fitrms = -999.0

    def set_command(self, command: Command[JaegerActor]):
        """Sets the command."""

        if command.status.is_done:
            raise FVCError("Command is done.")

        self.command = command

    def log(
        self,
        msg: str,
        level: logging._Level = logging.INFO,
        to_log: bool = True,
        to_command: bool = True,
    ):
        """Logs a message, including to the command if present."""

        level = int(level)
        msg = "[FVC]: " + msg

        if log and to_log:
            log.log(level, msg)

        if self.command and to_command:
            if level == logging.DEBUG:
                self.command.debug(msg)
            elif level == logging.INFO:
                self.command.info(msg)
            elif level == logging.WARNING:
                self.command.warning(msg)
            elif level == logging.ERROR:
                self.command.error(msg)

    async def expose(
        self,
        exposure_time: float = 5.0,
        stack: int = 1,
        use_tron_fallback=True,
    ) -> pathlib.Path:  # pragma: no cover
        """Takes an exposure with the FVC and blocks until the exposure is complete.

        Returns the path to the new image. If ``use_tron_fallback=True`` and the
        command has not been set, creates a Tron client to command the FVC.

        """

        if self.command is None:
            if use_tron_fallback is False:
                raise FVCError("Command must be set.")

        else:
            if self.command.status.is_done:
                raise FVCError("Command is done.")

        self.log(f"Taking {exposure_time} seconds FVC exposure.", to_command=False)

        tron = None

        if stack <= 1:
            cmd_str = f"talk -c fvc expose {exposure_time}"
        else:
            cmd_str = f"talk -c fvc expose --stack {stack} {exposure_time}"

        if self.command:
            expose_command = self.command.send_command("fliswarm", cmd_str)
        else:
            tron = TronConnection(
                config["actor"]["tron_host"],
                config["actor"]["tron_port"],
            )
            await tron.start()
            expose_command = tron.send_command("fliswarm", cmd_str)

        assert isinstance(expose_command, Command)
        await expose_command

        if tron:
            tron.stop()

        if expose_command.status.did_fail:
            raise FVCError("The FVC exposure failed.")

        for reply in expose_command.replies:
            for keyword in reply.keywords:
                if keyword.name.lower() == "filename":
                    filename = keyword.values[-1]
                    self.log(f"FVC raw image is {filename}.", to_command=False)
                    return pathlib.Path(filename)

        raise FVCError("The exposure succeeded but did not output the filename.")

    def process_fvc_image(
        self,
        path: pathlib.Path | str,
        fibre_data: Optional[pandas.DataFrame] = None,
        fibre_type: str = "Metrology",
        plot: bool | str = False,
        polids: numpy.ndarray | list | None = None,
    ) -> tuple[fits.ImageHDU, pandas.DataFrame, pandas.DataFrame]:
        """Processes a raw FVC image.

        Parameters
        ----------
        path
            The path to the raw FVC image.
        fibre_data
            A Pandas data frame with the expected coordinates of the targets. It
            is expected the data frame will have columns ``positioner_id``,
            ``hole_id``, ``fibre_type``, ``xwok``, and ``ywok``. This frame is
            appended to the processed image. Normally this parameters is left
            empty and the fibre table from the configuration loaded into the FPS
            instace is used.
        fibre_type
            The ``fibre_type`` rows in ``fibre_data`` to use. Defaults to
            ``fibre_type='Metrology'``.
        plot
            Whether to save additional debugging plots along with the processed image.
            If ``plot`` is a string, it will be used as the directory to which to
            save the plots.

        Returns
        -------
        result
            A tuple with the read raw image HDU (with columns flipped) as the first
            argument and the expected coordinates, as a data frame, as the second.
            The data frame is the same as the input target coordinates but with the
            columns ``xwok_measured`` and ``ywok_measured`` appended.

        """

        # Reset the instance
        self.reset()

        path = str(path)
        if not os.path.exists(path):
            raise FVCError(f"FVC image {path} does not exist.")

        if fibre_data is None:
            if self.fps is None or self.fps.configuration is None:
                raise FVCError("No fibre data and no configuration has been loaded.")
            fibre_data = self.fps.configuration.assignment_data.fibre_table

        self.fibre_data = fibre_data.copy().reset_index().set_index("fibre_type")

        self.log(f"Processing raw image {path}")

        dirname, base = os.path.split(path)
        proc_path_root = os.path.join(dirname, "proc-" + base[0 : base.find(".fit")])

        if plot is True:
            plot_path_root = proc_path_root
        elif isinstance(plot, str):
            plot_path_root = os.path.join(plot, "proc-" + base[0 : base.find(".fit")])
        else:
            plot_path_root = ""

        hdus = fits.open(path)

        self.image_path = path
        self.raw_hdu = hdus[1].copy()

        # Invert columns
        hdus[1].data = hdus[1].data[:, ::-1]
        image_data = hdus[1].data

        self.log(f"Max counts in image: {numpy.max(image_data)}", level=logging.INFO)

        self.centroids = self.extract(image_data)

        fiducialCoords = calibration.fiducialCoords.loc[self.site]

        xCMM = fiducialCoords.xWok.to_numpy()
        yCMM = fiducialCoords.yWok.to_numpy()
        xyCMM = numpy.array([xCMM, yCMM]).T

        xyCCD = self.centroids[["x", "y"]].to_numpy()

        fibre_data_met = self.fibre_data.loc[fibre_type]

        # Get close enough to associate the correct centroid with the correct fiducial.
        x_wok_expect = numpy.concatenate([xCMM, fibre_data_met.xwok.to_numpy()])
        y_wok_expect = numpy.concatenate([yCMM, fibre_data_met.ywok.to_numpy()])
        xy_wok_expect = numpy.array([x_wok_expect, y_wok_expect]).T

        rt = RoughTransform(xyCCD, xy_wok_expect)
        xy_wok_rough = rt.apply(xyCCD)

        # First associate fiducials and build first round just use outer fiducials
        rCMM = numpy.sqrt(xyCMM[:, 0] ** 2 + xyCMM[:, 1] ** 2)
        keep = rCMM > 310
        xyCMMouter = xyCMM[keep, :]

        arg_found, fid_rough_dist = arg_nearest_neighbor(xyCMMouter, xy_wok_rough)
        self.log(
            f"Max fiducial rough distance: {numpy.max(fid_rough_dist):.3f}",
            level=logging.INFO,
        )

        xy_fiducial_CCD = xyCCD[arg_found]
        xy_fiducial_wok_rough = xy_wok_rough[arg_found]

        if plot:
            self.plot_fvc_assignments(
                xy_wok_rough,
                fibre_data_met,
                xCMM,
                yCMM,
                plot_path_root + "_roughassoc.pdf",
                xy_fiducial=xy_fiducial_wok_rough,
                xy_fiducial_cmm=xyCMMouter,
                title="Rough fiducial association",
            )

        ft = ZhaoBurgeTransform(
            xy_fiducial_CCD,
            xyCMMouter,
            polids=(polids or config["fvc"]["zb_polids"]),
        )
        self.log(
            f"Full transform 1. Bisased RMS={ft.rms * 1000:.3f}, "
            f"Unbiased RMS={ft.unbiasedRMS * 1000:.3f}.",
            level=logging.INFO,
        )
        xy_wok_meas = ft.apply(xyCCD, zb=False)

        if plot:
            self.plot_fvc_assignments(
                xy_wok_meas,
                fibre_data_met,
                xCMM,
                yCMM,
                plot_path_root + "_full1.pdf",
                title="Full transform 1",
            )

        # Re-associate fiducials, some could have been wrongly associated in
        # first fit but second fit should be better?
        arg_found, fid_rough_dist = arg_nearest_neighbor(xyCMM, xy_wok_meas)
        self.log(
            f"Max fiducial fit 2 distance: {numpy.max(fid_rough_dist):.3f}",
            level=logging.INFO,
        )

        xy_fiducial_CCD = xyCCD[arg_found]  # Overwriting
        xy_fiducial_wok_refine = xy_wok_meas[arg_found]

        if plot:
            self.plot_fvc_assignments(
                xy_wok_meas,
                fibre_data_met,
                xCMM,
                yCMM,
                plot_path_root + "_refineassoc.pdf",
                title="Refined fiducial association",
                xy_fiducial=xy_fiducial_wok_refine,
                xy_fiducial_cmm=xyCMM,
            )

        # Try a new transform
        ft = ZhaoBurgeTransform(
            xy_fiducial_CCD,
            xyCMM,
            polids=(polids or config["fvc"]["zb_polids"]),
        )
        self.log(
            f"Full transform 2. Bisased RMS={ft.rms * 1000:.3f}, "
            f"Unbiased RMS={ft.unbiasedRMS * 1000:.3f}.",
            level=logging.INFO,
        )

        xy_wok_meas = ft.apply(xyCCD)  # Overwrite

        if plot:
            self.plot_fvc_assignments(
                xy_wok_meas,
                fibre_data_met,
                xCMM,
                yCMM,
                plot_path_root + "_full2.pdf",
                title="Full transform 2",
            )

        # Transform all CCD detections to wok space
        xy_expect_pos = fibre_data_met[["xwok", "ywok"]].to_numpy()

        arg_found, met_dist = arg_nearest_neighbor(xy_expect_pos, xy_wok_meas)
        self.log(
            f"Max metrology distance: {numpy.max(met_dist):.3f}",
            level=logging.INFO,
        )
        xy_wok_robot_meas = xy_wok_meas[arg_found]

        self.fibre_data.loc[fibre_type, "xwok_measured"] = xy_wok_robot_meas[:, 0]
        self.fibre_data.loc[fibre_type, "ywok_measured"] = xy_wok_robot_meas[:, 1]

        # Only use online robots for final RMS.
        online = self.fibre_data.loc[
            (self.fibre_data.index == fibre_type) & (self.fibre_data.offline == 0)
        ]
        dx = online.xwok - online.xwok_measured
        dy = online.ywok - online.ywok_measured

        self.fitrms = numpy.sqrt(numpy.mean(dx ** 2 + dy ** 2))
        self.log(f"RMS full fit {self.fitrms * 1000:.3f} um.")

        hdus[1].header["FITRMS"] = (self.fitrms * 1000, "RMS full fit [um]")

        self.fibre_data.reset_index(inplace=True)
        self.fibre_data.set_index(["hole_id", "fibre_type"], inplace=True)
        self.proc_hdu = hdus[1]

        self.log(f"Finished processing {path}", level=logging.DEBUG)

        return (self.proc_hdu, self.fibre_data, self.centroids)

    def calculate_offsets(
        self,
        reported_positions: numpy.ndarray,
        fibre_data: Optional[pandas.DataFrame] = None,
        k: Optional[float] = None,
    ) -> pandas.DataFrame:
        """Determines the offset to apply to the currently reported positions.

        Measured wok positions from the fibre data are converted to positioner
        coordinates. An alpha/beta offset is calculated with respect to the
        expected positions. The offsets is then applied to the current positions
        as self-reported by the positioners. Optionally, the offset can be
        adjusted using a PID loop.

        Parameters
        ----------
        reported_positions
            Reported positions for the positioners as a numpy array. Usually
            the output of `.FPS.get_positions`.
        fibre_data
            The fibre data table. Only the metrology entries are used. Must
            have the ``xwok_measured`` and ``ywok_measured`` column populated.
            If `None`, uses the data frame calculated when `.process_fvc_image`
            last run.
        k
            The fraction of the correction to apply.

        Returns
        -------
        new_positions
            The new alpha and beta positions as a Pandas dataframe indexed by
            positions ID. If `None`, uses the value ``fvc.k`` from the configuration.

        """

        self.log("Calculating offset from FVC image and fit.")

        site = config["observatory"]
        self.k = k or config["fvc"]["k"]

        if fibre_data is None and self.fibre_data is None:
            raise FVCError("No fibre data passed or stored in the instance.")

        if fibre_data is None:
            fibre_data = self.fibre_data
            assert fibre_data is not None

        fibre_data = fibre_data.copy().reset_index()
        met: pandas.DataFrame = fibre_data.loc[fibre_data.fibre_type == "Metrology"]

        # TODO: deal with missing data
        if (met.loc[:, ["xwok_measured", "ywok_measured"]] == -999.0).any().any():
            raise FVCError("Some metrology fibres have not been measured.")

        # Calculate alpha/beta from measured wok coordinates.
        _measured = []
        first = True
        for _, row in met.iterrows():
            (alpha_measured, beta_measured), _ = wok_to_positioner(
                row.hole_id,
                site,
                "Metrology",
                row.xwok_measured,
                row.ywok_measured,
            )

            if "alpha" in row:
                alpha_expected = row.alpha
                beta_expected = row.beta
            else:
                if first:
                    self.log(
                        "Fibre data does not include the expected alpha/beta "
                        "positions. Using reported alpha/beta.",
                        logging.WARNING,
                    )
                prow = reported_positions[reported_positions[:, 0] == row.positioner_id]
                alpha_expected = prow[0][1]
                beta_expected = prow[0][2]

            # If beta >= 180, we would need a left handed configuration. For now we
            # invalidate these values.
            if beta_expected >= 180.0:
                alpha_measured = numpy.nan
                beta_measured = numpy.nan

            xwok_distance = row.xwok_measured - row.xwok
            ywok_distance = row.ywok_measured - row.ywok

            _measured.append(
                (
                    row.hole_id,
                    row.positioner_id,
                    xwok_distance,
                    ywok_distance,
                    alpha_expected,
                    beta_expected,
                    alpha_measured,
                    beta_measured,
                )
            )

            first = False

        measured = pandas.DataFrame(
            _measured,
            columns=[
                "hole_id",
                "positioner_id",
                "xwok_distance",
                "ywok_distance",
                "alpha_expected",
                "beta_expected",
                "alpha_measured",
                "beta_measured",
            ],
        )
        measured.set_index("positioner_id", inplace=True)

        # Merge the reported positions.
        reported = pandas.DataFrame(
            reported_positions,
            columns=["positioner_id", "alpha_reported", "beta_reported"],
        )
        reported.positioner_id = reported.positioner_id.astype("int32")
        reported.set_index("positioner_id", inplace=True)

        offsets = pandas.concat([measured, reported], axis=1)

        # If there are measured alpha/beta that are NaN, replace those with the
        # previous value.
        offsets["transformation_valid"] = 1
        pos_na = offsets["alpha_measured"].isna()
        if pos_na.sum() > 0:
            self.log(
                "Failed to calculated corrected positioner coordinates for "
                f"{pos_na.sum()} positioners.",
                level=logging.WARNING,
            )

            # For now set these values to the expected because we'll use them to
            # calculate the offset (which will be zero for invalid conversions).
            expected = offsets.loc[pos_na, ["alpha_expected", "beta_expected"]]
            offsets.loc[pos_na, ["alpha_measured", "beta_measured"]] = expected
            offsets.loc[pos_na, "transformation_valid"] = 0

        # Calculate offset between expected and measured.
        alpha_offset = offsets["alpha_expected"] - offsets["alpha_measured"]
        beta_offset = offsets["beta_expected"] - offsets["beta_measured"]

        offsets["alpha_offset"] = alpha_offset
        offsets["beta_offset"] = beta_offset

        # Clip very large offsets and apply a proportional term.
        max_offset = config["fvc"]["max_offset"]
        alpha_offset_c = numpy.clip(self.k * alpha_offset, -max_offset, max_offset)
        beta_offset_c = numpy.clip(self.k * beta_offset, -max_offset, max_offset)

        offsets["alpha_offset_corrected"] = alpha_offset_c
        offsets["beta_offset_corrected"] = beta_offset_c

        # Calculate new alpha and beta by applying the offset to the reported
        # positions (i.e., the ones the positioners believe they are at).
        alpha_new = offsets["alpha_reported"] + offsets["alpha_offset_corrected"]
        beta_new = offsets["beta_reported"] + offsets["beta_offset_corrected"]

        # Get new positions that are out of range and clip them.
        alpha_oor = (alpha_new < 0.0) | (alpha_new > 360.0)
        beta_oor = (beta_new < 0.0) | (beta_new > 180.0)

        alpha_new.loc[alpha_oor] = numpy.clip(alpha_new.loc[alpha_oor], 0, 360)
        beta_new.loc[beta_oor] = numpy.clip(beta_new.loc[beta_oor], 0, 180)

        # For the values out of range, recompute the corrected offsets.
        alpha_corr = alpha_new.loc[alpha_oor] - offsets.loc[alpha_oor, "alpha_reported"]
        offsets.loc[alpha_oor, "alpha_offset_corrected"] = alpha_corr
        beta_corr = beta_new.loc[beta_oor] - offsets.loc[beta_oor, "beta_reported"]
        offsets.loc[beta_oor, "beta_offset_corrected"] = beta_corr

        # Final check. If alpha/beta_new are NaNs, replace with reported values.
        alpha_new[numpy.isnan(alpha_new)] = offsets.loc[
            numpy.isnan(alpha_new), "alpha_reported"
        ]
        beta_new[numpy.isnan(beta_new)] = offsets.loc[
            numpy.isnan(beta_new), "beta_reported"
        ]

        # Save new positions.
        offsets["alpha_new"] = alpha_new
        offsets["beta_new"] = beta_new

        # Set the invalid alpha/beta_measured back to NaN.
        invalid = offsets["transformation_valid"] == 0
        offsets.loc[
            invalid,
            [
                "alpha_measured",
                "beta_measured",
                "alpha_offset",
                "beta_offset",
                "alpha_offset_corrected",
                "beta_offset_corrected",
            ],
        ] = numpy.nan

        self.offsets = offsets

        self.log("Finished calculating offsets.", level=logging.DEBUG)

        return offsets

    async def write_proc_image(
        self,
        new_filename: Optional[str | pathlib.Path] = None,
    ) -> fits.HDUList:  # pragma: no cover
        """Writes the processed image along with additional table data.

        If ``new_filename`` is not passed, defaults to adding the prefix ``proc-``
        to the last processed image file path.

        """

        if self.image_path is None or self.proc_hdu is None:
            raise FVCError(
                "No current image. Take and process "
                "an image before callin write_proc_image()."
            )

        if new_filename is None:
            image_path = pathlib.Path(self.image_path)
            new_filename = image_path.with_name("proc-" + image_path.name)

        if (
            self.fibre_data is None
            or self.centroids is None
            or self.raw_hdu is None
            or self.proc_hdu is None
        ):
            raise FVCError("Need to run process_fvc_image before writing the image.")

        proc_hdus = fits.HDUList([fits.PrimaryHDU(), self.proc_hdu])

        positionerTable = calibration.positionerTable
        wokCoords = calibration.wokCoords
        fiducialCoords = calibration.fiducialCoords

        dfs = [
            ("POSITIONERTABLE", positionerTable.reset_index()),
            ("WOKCOORDS", wokCoords.reset_index()),
            ("FIDUCIALCOORDS", fiducialCoords.reset_index()),
        ]

        for name, df in dfs:
            rec = Table.from_pandas(df).as_array()
            table = fits.BinTableHDU(rec, name=name)
            proc_hdus.append(table)

        measured_coords = self.fibre_data.copy()
        measured_coords.reset_index(inplace=True)
        measured_coords.sort_values("positioner_id", inplace=True)
        measured_coords_rec = Table.from_pandas(measured_coords).as_array()
        proc_hdus.append(fits.BinTableHDU(measured_coords_rec, name="FIBERDATA"))

        # Add IEB information
        ieb_keys = config["fvc"]["ieb_keys"]
        ieb_data = {key: -999.0 for key in ieb_keys}

        for key in ieb_keys:
            device_name = ieb_keys[key]
            if self.fps and isinstance(self.fps.ieb, IEB):
                try:
                    device = self.fps.ieb.get_device(device_name)
                    ieb_data[key] = (await device.read())[0]
                except Exception as err:
                    self.log(f"Failed getting IEB information: {err}", logging.WARNING)
                    break

        for key, val in ieb_data.items():
            proc_hdus[1].header[key] = val

        if self.fps:
            await self.fps.update_position()
            positions = self.fps.get_positions()
            current_positions = pandas.DataFrame(
                {
                    "positionerID": positions[:, 0].astype(int),
                    "alphaReport": positions[:, 1],
                    "betaReport": positions[:, 2],
                }
            )

            current_positions["cmdAlpha"] = numpy.nan
            current_positions["cmdBeta"] = numpy.nan
            current_positions["startAlpha"] = numpy.nan
            current_positions["startBeta"] = numpy.nan

            if self.fps.configuration:
                robot_grid = self.fps.configuration.robot_grid

                _cmd_alpha = []
                _cmd_beta = []
                _start_alpha = []
                _start_beta = []

                if len(list(robot_grid.robotDict.values())[0].alphaPath) > 0:
                    for pid in current_positions.positionerID:
                        robot = robot_grid.robotDict[pid]
                        _cmd_alpha.append(robot.alphaPath[0][1])
                        _cmd_beta.append(robot.betaPath[0][1])
                        _start_alpha.append(robot.alphaPath[-1][1])
                        _start_beta.append(robot.betaPath[-1][1])

                    current_positions["cmdAlpha"] = _cmd_alpha
                    current_positions["cmdBeta"] = _cmd_beta
                    current_positions["startAlpha"] = _start_alpha
                    current_positions["startBeta"] = _start_beta

            rec = Table.from_pandas(current_positions.reset_index()).as_array()
            proc_hdus.append(fits.BinTableHDU(rec, name="POSANGLES"))

        rec = Table.from_pandas(self.centroids.reset_index()).as_array()
        proc_hdus.append(fits.BinTableHDU(rec, name="CENTROIDS"))

        if self.offsets is not None:
            rec = Table.from_pandas(self.offsets.reset_index()).as_array()
            proc_hdus.append(fits.BinTableHDU(rec, name="OFFSETS"))

        await run_in_executor(proc_hdus.writeto, new_filename, checksum=True)

        self.log(f"Processed HDU written to {new_filename}")

        return proc_hdus

    async def apply_correction(
        self,
        offsets: Optional[pandas.DataFrame] = None,
    ):  # pragma: no cover
        """Applies the offsets. Fails if the trajectory is collided or deadlock."""

        self.log("Preparing correction trajectory.")

        if self.offsets is None and offsets is None:
            raise FVCError("Offsets not set or passed. Cannot apply correction.")

        if offsets is None:
            offsets = self.offsets

        assert offsets is not None
        await self.fps.update_position()

        # Setup robot grid.
        grid = get_robot_grid()
        for robot in grid.robotDict.values():
            positioner = self.fps[robot.id]
            robot.setAlphaBeta(positioner.alpha, positioner.beta)

            row = offsets.loc[robot.id, ["alpha_new", "beta_new"]]
            if row.isna().any():
                log.warning(f"Positioner {robot.id}: new position is NaN. Skipping.")
                robot.setDestinationAlphaBeta(positioner.alpha, positioner.beta)
            else:
                robot.setDestinationAlphaBeta(row.alpha_new, row.beta_new)

        # Check for collisions.
        collided = [rid for rid in grid.robotDict if grid.isCollided(rid)]
        n_coll = len(collided)
        if n_coll > 0:
            raise FVCError(f"Cannot apply corrections. {n_coll} robots are collided.")

        # Generate trajectories.
        (to_destination, _, did_fail, deadlocks) = await get_path_pair_in_executor(
            grid,
            ignore_did_fail=True,
        )
        if did_fail:
            log.warning(
                f"Found {len(deadlocks)} deadlocks but applying correction anyway."
            )

        self.log("Sending correction trajectory.")
        try:
            await self.fps.send_trajectory(to_destination)
        except TrajectoryError as err:
            raise FVCError(f"Failed executing the correction trajectory: {err}")

        self.log("Correction applied.")

    def extract(self, image_data: numpy.ndarray) -> pandas.DataFrame:
        """Extract image data using SExtractor. Returns the extracted centroids."""

        image_data = numpy.array(image_data, dtype=numpy.float32)

        bkg = sep.Background(image_data)
        bkg_image = bkg.back()

        data_sub = image_data - bkg_image

        objects = sep.extract(
            data_sub,
            config["fvc"]["background_sigma"],
            err=bkg.globalrms,
        )
        objects = pandas.DataFrame(objects)

        # Eccentricity
        objects["ecentricity"] = 1 - objects["b"] / objects["a"]

        # Slope of ellipse (optical distortion direction)
        objects["slope"] = numpy.tan(objects["theta"] + numpy.pi / 2)  # rotate by 90

        # Intercept of optical distortion direction
        objects["intercept"] = objects["y"] - objects["slope"] * objects["x"]

        # Ignore everything less than X pixels
        objects = objects.loc[objects["npix"] > config["fvc"]["centroid_min_npix"]]

        self.log(f"Found {len(objects)} centroids", level=logging.INFO)

        ncentroids = len(calibration.positionerTable) + len(calibration.fiducialCoords)
        self.log(f"Expected {ncentroids} centroids", level=logging.INFO)

        return objects

    def plot_fvc_assignments(
        self,
        xy: numpy.ndarray,
        target_coords: pandas.DataFrame | pandas.Series,
        xCMM: numpy.ndarray,
        yCMM: numpy.ndarray,
        filename: str,
        xy_fiducial: numpy.ndarray | None = None,
        xy_fiducial_cmm: numpy.ndarray | None = None,
        title: str | None = None,
    ):
        """Plot the results of the transformation."""

        plt.figure(figsize=(8, 8))

        if title:
            plt.title(title)

        plt.plot(
            xy[:, 0],
            xy[:, 1],
            "o",
            ms=4,
            markerfacecolor="None",
            markeredgecolor="red",
            markeredgewidth=1,
            label="Centroid",
        )

        plt.plot(
            target_coords.xwok.to_numpy(),
            target_coords.ywok.to_numpy(),
            "xk",
            ms=3,
            label="Expected MET",
        )

        # Overplot fiducials
        plt.plot(
            xCMM,
            yCMM,
            "D",
            ms=6,
            markerfacecolor="None",
            markeredgecolor="cornflowerblue",
            markeredgewidth=1,
            label="Expected FIF",
        )

        if xy_fiducial is not None and xy_fiducial_cmm is not None:
            for cmm, measured in zip(xy_fiducial_cmm, xy_fiducial):
                plt.plot([cmm[0], measured[0]], [cmm[1], measured[1]], "-k")

        plt.axis("equal")
        plt.legend()
        plt.xlim([-350, 350])
        plt.ylim([-350, 350])
        plt.savefig(filename, dpi=350)
        plt.close()


def arg_nearest_neighbor(xyA: numpy.ndarray, xyB: numpy.ndarray):
    """Loop over xy list A, find nearest neighbour in list B

    Returns
    -------
    result
        The indices in list b that match A.

    """

    # TODO: this is probably efficient given the number of points, but maybe
    # replace with a scipy cdist.

    xyA = numpy.array(xyA)
    xyB = numpy.array(xyB)
    out = []
    distance = []
    for x, y in xyA:
        dist = numpy.sqrt((x - xyB[:, 0]) ** 2 + (y - xyB[:, 1]) ** 2)
        amin = numpy.argmin(dist)
        distance.append(dist[amin])
        out.append(amin)

    return numpy.array(out), numpy.array(distance)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-05-12
# @Filename: ieb.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import os

from typing import Any, Dict

from drift import Drift, DriftError

from jaeger import config


__all__ = ["IEB", "FVC"]


class IEB(Drift):
    """Thing wrapper around a :class:`~drift.drift.Drift` class.

    Allows additional features such as disabling the interface.

    """

    def __init__(self, *args, **kwargs):

        self.disabled = False

        super().__init__(*args, **kwargs)

        self._categories = None

    @classmethod
    def create(cls, path=None):
        """Creates an `.IEB` instance with the default configuration."""

        default_ieb_path = path or config["files"]["ieb_config"]

        default_ieb_path = os.path.expanduser(os.path.expandvars(default_ieb_path))
        if not os.path.isabs(default_ieb_path):
            default_ieb_path = os.path.join(os.path.dirname(__file__), default_ieb_path)

        return cls.from_config(default_ieb_path)

    def get_categories(self):
        """Returns a list of categories."""

        if self._categories is None:
            categories = [
                device.category
                for module in self.modules
                for device in self.modules[module].devices.values()
            ]
            self._categories = set(categories)
            self._categories.discard(None)

        return self._categories

    async def __aenter__(self):

        if self.disabled:
            raise DriftError("IEB is disabled.")

        try:
            await Drift.__aenter__(self)
        except DriftError:
            self.disabled = True
            raise DriftError("Failed connecting to the IEB. Disabling it.")

    async def __aexit__(self, *args):

        await Drift.__aexit__(self, *args)

    async def get_status(self) -> Dict[str, Any]:
        """Returns the status of the IEB components."""

        status = {}
        for category in self.get_categories():
            data = await self.read_category(category)
            for device in data:
                if self.get_device(device).__type__ == "relay":
                    value = False if data[device][0] == "open" else True
                else:
                    value = data[device][0]
                status[device] = value

        return status


class FVC(IEB):
    """Connects to the FVC IEB."""

    @classmethod
    def create(cls, path=None):
        """Creates an `.FVC` instance with the default configuration."""

        default_ieb_path = config["files"]["fvc_config"]

        return super().create(default_ieb_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-10-12
# @Filename: database.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from functools import wraps
from glob import glob

import pandas
import peewee
from astropy import table
from astropy.io import fits

from sdssdb.peewee.sdss5db import targetdb


__all__ = ["load_holes", "load_fields"]


def check_database(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if targetdb.database.connected is False:
            raise RuntimeError("Database is not connected.")
        return f(*args, **kwargs)

    return wrapper


@check_database
def load_holes(observatory: str):
    """Loads a list holes to ``targetdb.hole``."""

    observatory_pk = targetdb.Observatory.get(label=observatory).pk

    row_start = 13
    row_end = -13
    min_cols = 14

    holes = []
    for row in range(row_start, row_end - 1, -1):
        end_col = min_cols + ((row_start - row) if row >= 0 else (row - row_end))
        for col in range(1, end_col + 1, 1):
            if row == 0:
                holeid = f"R0C{col}"
            else:
                holeid = f"R{row:+}C{col}"

            holes.append(
                dict(
                    row=row,
                    column=col,
                    holeid=holeid,
                    observatory_pk=observatory_pk,
                )
            )

    targetdb.Hole.insert(holes).on_conflict(
        conflict_target=[targetdb.Hole.holeid, targetdb.Hole.observatory],
        action="IGNORE",
    ).execute(targetdb.database)


@check_database
def load_fields(plan: str, files: list[str] = None, pattern: str = None):
    """Loads a series of field.

    Parameters
    ----------
    plan
        The robostrategy run string.
    files
        A list of files to load.
    pattern
        Alternative to ``files``, a pattern to be used with ``glob``
        to retrieve a list of files to load.
    """

    if files is None and pattern:
        files = list(glob(pattern))

    if files is None or len(files) == 0:
        raise ValueError("No files provided.")

    version = targetdb.Version.get_or_create(
        plan=plan,
        target_selection=False,
        robostrategy=True,
    )

    hole_ids = pandas.DataFrame(
        list(
            targetdb.Hole.select(
                targetdb.Hole.pk,
                targetdb.Hole.holeid,
                targetdb.Observatory.label,
            )
            .join(targetdb.Observatory)
            .where(targetdb.Hole.holeid.is_null(False))
            .tuples()
        ),
        columns=["pk", "holeid", "observatory"],
    ).set_index("holeid")

    for file_ in files:
        hdul = fits.open(file_)

        # Create field
        field_id = int(file_.split("-")[-1][:-5])
        observatory = hdul[0].header["OBS"].upper()
        racen = hdul[0].header["RACEN"]
        deccen = hdul[0].header["DECCEN"]
        PA = hdul[0].header["PA"]
        field_cadence = hdul[0].header["FCADENCE"]
        nexp = hdul[0].header["NEXP"]

        insert = targetdb.Field.insert(
            field_id=field_id,
            racen=racen,
            deccen=deccen,
            position_angle=PA,
            version_pk=version[0].pk,
            cadence_pk=targetdb.Cadence.get(label=field_cadence).pk,
            observatory=targetdb.Observatory.get(label=observatory).pk,
        ).on_conflict(
            conflict_target=[targetdb.Field.field_id],
            preserve=[targetdb.Field.field_id],
            update={
                "observatory_pk": peewee.EXCLUDED.observatory_pk,
                "racen": peewee.EXCLUDED.racen,
                "deccen": peewee.EXCLUDED.deccen,
                "position_angle": peewee.EXCLUDED.position_angle,
                "version_pk": peewee.EXCLUDED.version_pk,
                "cadence_pk": peewee.EXCLUDED.cadence_pk,
            },
        )
        field_pk = insert.execute(targetdb.database)

        # Now create a design for each exposure.
        design_modes = hdul[0].header["DESMODE"].split()

        assign_data = table.hstack(
            (
                table.Table(hdul["ASSIGN"].data),
                table.Table(hdul["TARGET"].data),
            )
        )
        assign_data = assign_data[assign_data["assigned"] > 0]
        insert_data = []

        for n in range(nexp):
            design = targetdb.Design(
                exposure=n + 1,
                field_pk=field_pk,
                design_mode_label=design_modes[n],
            )
            design.save()

            if nexp == 1:
                exp_data = assign_data
                holeIDs = exp_data["holeID"].tolist()
            else:
                exp_data = assign_data[assign_data["holeID"][:, n] != " "]
                holeIDs = exp_data["holeID"][:, n].tolist()

            obs_holes = hole_ids.loc[hole_ids.observatory == observatory]
            holeid_pk_exp = obs_holes.loc[holeIDs].pk.tolist()

            insert_data += [
                {
                    "design_id": design.design_id,
                    "hole_pk": holeid_pk_exp[i],
                    "carton_to_target_pk": exp_data["carton_to_target_pk"][i],
                    "instrument_pk": targetdb.Instrument.get(
                        label=exp_data["fiberType_2"][i]
                    ).pk,
                }
                for i in range(len(exp_data))
            ]

        targetdb.Assignment.insert_many(insert_data).execute(targetdb.database)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-11-12
# @Filename: ieb.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio
import math

from typing import Tuple

import click

from clu.command import Command
from clu.parsers.click import pass_args

from jaeger.fps import FPS
from jaeger.ieb import IEB
from jaeger.testing import VirtualFPS

from .. import JaegerActor
from . import jaeger_parser


__all__ = ["ieb"]


async def _get_category_data(command, category) -> list:

    ieb = command.actor.fps.ieb
    schema = command.actor.model.schema

    items = schema["properties"][category]["items"]
    measured = []

    async with ieb:
        for item in items:
            name = item["title"]
            type_ = item["type"]
            device = ieb.get_device(name)
            value = (await device.read(connect=False))[0]
            if type_ == "boolean" and device.__type__ == "relay":
                value = True if value == "closed" else False
            elif type_ == "integer":
                value = int(value)
            elif type_ == "number":
                if "multipleOf" in item:
                    precision = int(-math.log10(item["multipleOf"]))
                else:
                    precision = 3
                value = round(value, precision)
            measured.append(value)

    return measured


@jaeger_parser.group()
@pass_args()
def ieb(command, fps):
    """Manages the IEB."""

    ieb = fps.ieb

    if not ieb or ieb.disabled:
        command.fail(error="ieb not connected.")
        raise click.Abort()

    return


@ieb.command()
async def enable(command, fps):
    """Re-enables the IEB."""

    if fps.ieb is None or not isinstance(fps.ieb, IEB):
        return command.fail("IEB object does not exist.")

    fps.ieb.disabled = False

    return command.finish()


@ieb.command()
async def status(command, fps):
    """Outputs the status of the devices."""

    categories = sorted(fps.ieb.get_categories())

    status_data = {}
    for category in categories:
        if category not in command.actor.model.schema["properties"]:
            command.warning(f"Unknown device category {category!r}.")
        status_data[category] = await _get_category_data(command, category)

    return command.finish(message=status_data, concatenate=True)


@ieb.command()
@click.argument("DEVICES", type=str, nargs=-1)
@click.option(
    "--on/--off",
    default=None,
    help="the desired state of the device. If not passed, switches the current state.",
)
@click.option(
    "--cycle",
    is_flag=True,
    help="power cycles a relay. The final status is on.",
)
@click.option(
    "--delay",
    type=float,
    default=1,
    help="When powering multiple devices, the delay to wait between them.",
)
async def switch(command, fps, devices=(), on=None, cycle=False, delay=1):
    """Switches the status of an on/off device."""

    ieb = fps.ieb

    if len(devices) == 0:
        return command.fail(error="No devices specified.")

    for idev, device in enumerate(devices):

        if len(devices) > 1 and idev != 0:
            await asyncio.sleep(delay)

        if cycle:
            on = False

        try:
            device_obj = ieb.get_device(device)
            dev_name = device_obj.name
            category = device_obj.category.lower()
        except ValueError:
            return command.fail(error=f"cannot find device {device!r}.")

        if device_obj.module.mode not in ["holding_register", "coil"]:
            return command.fail(error=f"{dev_name!r} is not an output.")

        if on is None:  # The --on/--off was not passed
            current_status = (await device_obj.read())[0]
            if current_status == "closed":
                on = False
            elif current_status == "open":
                on = True
            else:
                return command.fail(
                    error=f"invalid status for device {dev_name!r}: {current_status!r}."
                )

        try:
            if on is True:
                await device_obj.close()
            elif on is False:
                await device_obj.open()
        except Exception:
            return command.fail(error=f"failed to set status of device {dev_name!r}.")

        if cycle:
            command.write("d", text="waiting 1 second before powering up.")
            await asyncio.sleep(1)
            try:
                await device_obj.close()
            except Exception:
                return command.fail(
                    error=f"failed to power device {dev_name!r} back on."
                )

        # If we read the status immediately sometimes we still get the old one.
        # Sleep a bit to avoid that.
        await asyncio.sleep(0.2)

        status = "on" if (await device_obj.read())[0] == "closed" else "off"

        command.info(
            message={
                "text": f"device {dev_name!r} is now {status!r}.",
                category: await _get_category_data(command, category),
            }
        )

    return command.finish()


async def _power_sequence(command, ieb, seq, mode="on", delay=3) -> bool:
    """Applies the power on/off sequence."""

    # To speed up tests
    if isinstance(command.actor.fps, VirtualFPS):
        delay = 0.01

    relay_result = "closed" if mode == "on" else "open"

    command.info(text=f"Running power {mode} sequence")

    # First check that the SYNC line is open.
    sync = ieb.get_device("SYNC")
    if (await sync.read())[0] == "closed":
        command.debug(text="SYNC line is high. Opening it.")
        await sync.open()

    if (await sync.read())[0] != "open":
        command.fail(error="Failed opening SYNC line.")
        return False

    command.debug(power_sync=[False])

    for devname in seq:
        do_delay = False
        if isinstance(devname, str):
            dev = ieb.get_device(devname)
            category = dev.category.lower()

            if (await dev.read())[0] == relay_result:
                command.debug(text=f"{devname} already powered {mode}.")
            else:
                command.debug(text=f"Powering {mode} {devname}.")
                await dev.close() if mode == "on" else await dev.open()
                await asyncio.sleep(0.2)
                if (await dev.read())[0] != relay_result:
                    command.fail(error=f"Failed powering {mode} {devname}.")
                    return False

                do_delay = True

            command.debug({category: await _get_category_data(command, category)})

        elif isinstance(devname, (tuple, list)):
            devname = list(devname)
            devs = [ieb.get_device(dn) for dn in devname]
            category = devs[0].category.lower()

            status = list(await asyncio.gather(*[dev.read() for dev in devs]))

            keep = []
            for ii, res in enumerate(status):
                if res[0] == relay_result:
                    command.debug({"text": f"{devname[ii]} already powered {mode}."})
                    continue
                keep.append(ii)

            devs = [devs[ii] for ii in keep]
            devname = [devname[ii] for ii in keep]

            if len(devs) > 0:
                command.debug(text=f"Powering {mode} {', '.join(devname)}")
                await asyncio.gather(
                    *[dev.close() if mode == "on" else dev.open() for dev in devs]
                )
                do_delay = True

            status = list(await asyncio.gather(*[dev.read() for dev in devs]))
            for ii, res in enumerate(status):
                if res[0] != relay_result:
                    command.fail(error=f"Failed powering {mode} {devname[ii]}.")
                    return False

            command.debug({category: await _get_category_data(command, category)})

        else:
            command.fail(error=f"Invalid relay {devname!r}.")
            return False

        if do_delay:
            await asyncio.sleep(delay)

    return True


@ieb.group()
@pass_args()
def power(command, fps):
    """Runs the power on/off sequences."""

    pass


@power.command()
@click.option("--no-gfas", is_flag=True, help="Do not power the GFAs.")
async def on(command, fps: FPS, no_gfas: bool = False):
    """Powers on all the FPS IEB components."""

    command.info(text="Turning pollers off.")
    await fps.pollers.stop()

    ieb = fps.ieb
    if not isinstance(ieb, IEB) or ieb.disabled:
        return command.fail(error="IEB is not conencted or is disabled.")

    # Sequence of relays to power on. Tuples indicate relays that can be powered
    # on concurrently.
    on_seq = [
        ("CM1", "CM2", "CM3", "CM4", "CM5", "CM6"),
        "PS1",
        "PS2",
        "PS3",
        "PS4",
        "PS5",
        "PS6",
        "NUC1",
        "NUC2",
        "NUC3",
        "NUC4",
        "NUC5",
        "NUC6",
    ]

    if no_gfas is False:
        on_seq += [
            "GFA1",
            "GFA2",
            "GFA3",
            "GFA4",
            "GFA5",
            "GFA6",
        ]

    if not (await _power_sequence(command, ieb, on_seq, mode="on")):
        return

    return command.finish(text="Power on sequence complete.")


@power.command()
@click.option("--nucs", is_flag=True, help="Also power down NUCs")
async def off(command, fps, nucs):
    """Powers off all the FPS IEB components."""

    command.info(text="Turning pollers off.")
    await fps.pollers.stop()

    ieb = fps.ieb
    if not isinstance(ieb, IEB) or ieb.disabled:
        return command.fail(error="IEB is not conencted or is disabled.")

    # Sequence of relays to power off. Tuples indicate relays that can be powered
    # off concurrently.
    off_seq = [
        "PS1",
        "PS2",
        "PS3",
        "PS4",
        "PS5",
        "PS6",
        "GFA1",
        "GFA2",
        "GFA3",
        "GFA4",
        "GFA5",
        "GFA6",
    ]

    if nucs:
        off_seq += ["NUC1", "NUC2", "NUC3", "NUC4", "NUC5", "NUC6"]

    if not (await _power_sequence(command, ieb, off_seq, mode="off")):
        return

    return command.finish(text="Power off sequence complete.")


@ieb.command()
@click.argument("device_names", metavar="DEVICES", type=str, nargs=-1)
@click.argument("VALUE", type=click.FloatRange(0, 100))
async def fbi(command, fps: FPS, device_names: Tuple[str], value: float):
    """Control the power output (0-100%) of the fibre back illuminator (FBI)."""

    raw_value = 32 * int(1023 * (value / 100))

    if not isinstance(fps.ieb, IEB) or fps.ieb.disabled:
        return command.fail(error="IEB is not conencted or is disabled.")

    if len(device_names) == 0:
        return command.fail(error="No devices provided.")

    for device_name in device_names:
        try:
            device = fps.ieb.get_device(device_name)
        except ValueError:
            return command.fail(error=f"Cannot find device {device_name!r}.")

        if device.mode != "holding_register":
            return command.fail(
                error=f"Invalid device mode for {device_name!r}: {device.__type__}."
            )

        await device.write(raw_value)

    return command.finish()


@ieb.command()
@click.option("-v", "--verbose", is_flag=True, help="Shows extra information.")
async def info(command: Command[JaegerActor], fps: FPS, verbose=False):
    """Shows information about the IEB layout in a human-readable format."""

    ieb = fps.ieb

    if not isinstance(ieb, IEB):
        return command.fail(error="IEB is not conencted or is disabled.")

    modules = sorted(ieb.modules.keys())
    categories = set()

    command.info(text="Modules:")

    for module_name in modules:
        module = ieb.modules[module_name]

        command.info(text=f"  {module.name}:")

        if module.description != "":
            command.info(text=f"    description: {module.description}")

        if verbose:
            if module.model:
                command.info(text=f"    model: {module.model}")
            if module.mode:
                command.info(text=f"    mode: {module.mode}")
            if module.channels:
                command.info(text=f"    channels: {module.channels}")

        command.info(text="    Devices:")

        for dev in ieb[module_name].devices.values():
            command.info(text=f"      {dev.name}:")
            if dev.category != "":
                command.info(text=f"        category: {dev.category}")
                categories.add(dev.category)
            if dev.description != "":
                command.info(text=f"        description: {dev.description}")

            if verbose:
                if dev.address:
                    command.info(text=f"        address: {dev.address}")
                if dev.channel:
                    command.info(text=f"        channel: {dev.channel}")
                if dev.__type__:
                    command.info(text=f"        type: {dev.__type__}")
                    if dev.__type__ == "relay" and dev.relay_type:
                        command.info(text=f"        relay type: {dev.relay_type}")

    if len(categories) > 0 and command.actor and command.actor.model:

        command.info(text="")
        command.info(text="Keywords:")

        for category in sorted(categories):
            cat_data = command.actor.model.schema["properties"][category]
            items = [item["title"] for item in cat_data["items"]]
            command.info(text=f"  {category}=[{', '.join(items)}]")

    return command.finish()
